#include "config.h"
#include "genlib.h"
#include "stddefs.h"
#include "cli.h"
#include "xgpio.h"
#include "xgpio_l.h"
#include "xparameters.h"

#define XPAR_LDM_BASEADDR 0

/* This code is assuming the following bit organization for the 4-bits
 * of GPIO dedicated to LDM control:
 *
 * LDM_bit[0]:	(output)	SEN
 * LDM_bit[1]:	(output)	SCLK
 * LDM_bit[2]:	(output)	SDAT
 * LDM_bit[3]:	(input)		SDO
 */

#define SEN_BIT		0
#define SCLK_BIT	1
#define SDAT_BIT	2
#define SDO_BIT		3


void
SEN(int hi)
{
	Xuint32 val;

	val = XGpio_mReadReg(XPAR_LDM_BASEADDR,0);
	if (hi)
		XGpio_mWriteReg(XPAR_LDM_BASEADDR,0,val |  (1 << SEN_BIT))
	else
		XGpio_mWriteReg(XPAR_LDM_BASEADDR,0,val & ~(1 << SEN_BIT))
}

void
SCLK(int hi)
{
	Xuint32 val;

	val = XGpio_mReadReg(XPAR_LDM_BASEADDR,0);
	if (hi)
		XGpio_mWriteReg(XPAR_LDM_BASEADDR,0,val |  (1 << SCLK_BIT))
	else
		XGpio_mWriteReg(XPAR_LDM_BASEADDR,0,val & ~(1 << SCLK_BIT))
}

void
SDAT(int hi)
{
	Xuint32 val;

	val = XGpio_mReadReg(XPAR_LDM_BASEADDR,0);
	if (hi)
		XGpio_mWriteReg(XPAR_LDM_BASEADDR,0,val |  (1 << SDAT_BIT))
	else
		XGpio_mWriteReg(XPAR_LDM_BASEADDR,0,val & ~(1 << SDAT_BIT))
}

int
SDO(void)
{
	Xuint32 val;

	val = XGpio_mReadReg(XPAR_LDM_BASEADDR,0);
	if (val & (1 << SDO_BIT))
		return(1);
	return(0);
}

void
ldminit(void)
{
	Xuint32 val;

	
	XGpio_mSetDataDirection(XPAR_LEDS_8BIT_BASEADDR,0,0);
	XGpio_mSetDataDirection(XPAR_DIP_SWITCHES_8BIT_BASEADDR,0,0xff);
	XGpio_mSetDataDirection(XPAR_LDM_BASEADDR,0,0x08); /* 1=in, 0=out */

	while(!gotachar()) {
		val = XGpio_mReadReg(XPAR_DIP_SWITCHES_8BIT_BASEADDR,0);
		XGpio_mWriteReg(XPAR_LEDS_8BIT_BASEADDR,0,val);
	}
}

void
sendByte(uchar byte)
{
	uchar mask;

	for(mask = 0x80;mask >= 0x01;mask >>= 1) {
		SDAT(byte & mask);
		SCLK(1);
		monDelay(10);
		SCLK(0);
	}
}

uchar
recvByte(void)
{
	uchar mask, value;

	value = 0;
	for(mask = 0x80;mask >= 0x01;mask >>= 1) {
		SCLK(1);
		monDelay(10);
		SCLK(0);
		if (SDO())
			value |= mask;
	}
	return(value);
}

char *LdmHelp[] = {
	"Interface to LDM-0311-D1 display module",
	"-[irv] {addr} [value]",
#if INCLUDE_VERBOSEHELP
	"Options:",
	" -i   initialize gpio ports",
	" -r   read mode (default is write)",
	" -v   additive verbosity",
#endif
	0,
};

int
LdmCmd(int argc,char *argv[])
{
	uchar	addr, value;
	int		opt, verbose, read;

	read = verbose = 0;

	while((opt=getopt(argc,argv,"irv")) != -1) {
		switch(opt) {
		case 'i':
			ldminit();
			break;
		case 'r':
			read = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			return(CMD_PARAM_ERROR);
		}
	}

	if (argc < (optind+1))
		return(CMD_PARAM_ERROR);

	addr = (uchar)strtoul(argv[optind],0,0);
	if (addr > 0x7f)
		return(CMD_PARAM_ERROR);

	/* Set initial state:
	 */
	SEN(1);
	SCLK(0);

	if (read) {
		if (argc != (optind + 1))
			return(CMD_PARAM_ERROR);

		addr |= 0x80;

		SEN(0);
		sendByte(addr);			/* Send the address */
		value = recvByte();		/* Read the data */

		printf("Addr 0x%02x = 0x%02x\n",addr & 0x7f,value);
	}
	else {
		if (argc != (optind + 2))
			return(CMD_PARAM_ERROR);

		value = (uchar)strtoul(argv[optind+1],0,0);

		SEN(0);
		sendByte(addr);		/* Send the address */
		sendByte(value);	/* Send the data */
	}

	/* Set closing state:
	 */
	SCLK(0);
	SEN(1);

	return(CMD_SUCCESS);
}

