/* vi: set ts=4: */
/* bfspisd.c:
 * The blackfin-specific portion of the SPI-SD code...
 * Much of this code is based on text from the Simplified Physical Layer
 * Specification 2.00 (SPLS).
 *
 * In SPI mode, the SD card interface is byte oriented.  All communication
 * is controlled by the master's assertion of CS signal (low).
 *
 * According to section 3.3 (SPI Bus Topology) of the Sandisk SD-Card
 * Product Manual (version 1.9), the CS line is active low and must be
 * held low for the duration of the SPI transaction.  
 * This means that "normal" SPI mode of the BF537 can't be used because
 * it doesn't keep the SSEL line low across byte transfers.  As a result,
 * we have to control the SSEL line as a GPIO pin for this to work.
 *
 * In SPI mode, all data tokens are multiples of 8 bits and always byte
 * aligned to the CS signal.
 * 
 * Status (as of Dec 17, 2009):
 * Tested it with DOSFS:
 *  I had to undo the change submitted by Graham Henderson because
 *  with that change, I found that writing one file would remove the
 *  previously written file.  Have not confirmed this with Graham.
 * Tested with ELMFATFS (from http://elm-chan.org/fsw/ff/00index_e.html)
 *  Works well, and it also allowed me to test out the multiple block
 *  read/write (which apparently isn't used by DOSFS).
 *  Tested with ELMFATFS with 512Mg, 2Gig & 8Gig (high capacity) SD Cards.
 *  Found that numbers greater than 2G don't print (expected).  That's the
 *  only open issue I'm aware of at this point.
 *
 * Note:
 * This has not been tested in a multi-slave SPI configuration, so there may
 * be yet some work to be done for those configurations. Refer to the
 * section titled "Consideration on Multi-slave COnfiguration" on
 * http://elm-chan.org/docs/mmc/mmc_e.html for more information.  In the
 * nutshell, upon completion of a SD/SPI transaction it is necessary
 * to send one additional byte after chip-select is deasserted. 
 * This code does this inthe UNSELECT_SD() macro, but it is not tested.
 */
 
#include "config.h"
#if INCLUDE_SD
#include "cdefBF537.h"
#include "stddefs.h"
#include "genlib.h"
#include "cli.h"
#include "tfs.h"
#include "tfsprivate.h"
#include "timer.h"
#include "cpu.h"
#include "spi_bf537.h"
#include "sd.h"

/* Depending on which SPI select line is used, define these appropriately...
 */
#define SD_FLS	FLS5

#define CTLBITS	(MSTR | SPE)

#define SELECT_SD()		*pPORTFIO_CLEAR |= PF5
#define UNSELECT_SD()	{ *pPORTFIO_SET |= PF5; sdSpiWrite(0xff); }

uchar
sdSpiWrite(uchar byte) 
{ 
	*pSPI_STAT = TXCOL | RBSY | MODF | TXE; 
	*pSPI_CTL = CTLBITS | TDBR_CORE;

	*pSPI_TDBR = byte;
	__builtin_bfin_ssync();

	SPI_TXSWAIT();
	SPI_FINWAIT();

	return (uchar)*pSPI_RDBR; 
} 

/* sdPowerup():
 * Power up sequence; generate tot*8 clocks...
 * At startup, SPI-SD requires an initial 74 clocks prior to enabling chip
 * select, so this function provides that...
 */
int
sdPowerup(int tot)
{
	volatile int i;
	uchar tmp;

	*pSPI_STAT = TXCOL | RBSY | MODF | TXE; 
	__builtin_bfin_ssync();
	*pSPI_CTL = CTLBITS | RDBR_CORE;
	__builtin_bfin_ssync();
	tmp = (unsigned char)*pSPI_RDBR;
	__builtin_bfin_ssync();

	for(i=0;i<tot;i++) {
		SPI_RXSWAIT();
		__builtin_bfin_ssync();
		SPI_FINWAIT();
		__builtin_bfin_ssync();
		tmp = (unsigned char)*pSPI_RDBR;
		__builtin_bfin_ssync();
	}
	SPI_FINWAIT();
	return(0);
}

/* sdCardCmd():
 * Generic "send command" function for SPI/SD card interface.
 * All SD memory card commands are 6 bytes long.
 */
int
sdCardCmd(int interface, int cmd, ulong arg, uchar *resp)
{
	uchar buf[6];
	int i, rlen, rc = 0;
    struct elapsed_tmr tmr;

	if (CMD_APP(cmd))
		sdCardCmd(interface,APP_CMD,0,resp);

	rlen = sdCmdrlen(cmd);

	buf[0] = cmd | 0x40;
	buf[1] = (arg & 0xff000000) >> 24;
	buf[2] = (arg & 0xff0000) >> 16;
	buf[3] = (arg & 0xff00) >> 8;
	buf[4] = arg & 0xff;
	buf[5] = (cmd == 0 ? 0x95 : 0x01);	// CRC
	if (cmd == 8)
		buf[5] = crc7(0,buf,5);

	/* Don't know why I need to do this, but I do.  The sdPowerup() function
	 * was originally written to provide the initial 74-80 pre-chipselect
	 * clocks, but for some reason I need to call it here as well.
	 */
	sdPowerup(2);

	SPI_ENABLE();
	__builtin_bfin_ssync();

	SELECT_SD();
	__builtin_bfin_ssync();

	/* Send command:
	 */
	for(i=0;i<6;i++) 
		sdSpiWrite(buf[i]);

	/* Keep reading till the start bit is recieved (with timeout)...
	 * Once this is received, then read the remaining (if any)
	 * bytes of the response)...
	 */
    startElapsedTimer(&tmr,1000);
	while(1) {
		resp[0] = sdSpiWrite(0xff);
		if ((resp[0] & R1_STARTBIT) == 0)
			break;
    	if(msecElapsed(&tmr)) {
			if (sdVerbose & 2)
				printf("CMD_%d rd-timeout\n",cmd);
			rc = -1;
			goto done;
		}
	}

	/* Retrieve however many bytes are to be retrieved (based on
	 * the command type...
	 */
	for(i=1;i<rlen;i++) {	
		resp[i] = sdSpiWrite(0xff);
	}

	// Check for (and print) error in response...
	if ((sdVerbose & 2) && (resp[0] & R1_ERRMASK)) {
		printf("CMD_%d: ",cmd);
		if (resp[0] & R1_PARAM_ERR)
			printf("param err ");
		if (resp[0] & R1_ADDRESS_ERR)
			printf("address err ");
		if (resp[0] & R1_ERASESEQ_ERR)
			printf("eraseseq err ");
		if (resp[0] & R1_CRC_ERR)
			printf("crc err ");
		if (resp[0] & R1_ILLEGAL_CMD)
			printf("illegal cmd");
		putchar('\n');
	}

	SPI_FINWAIT();

done:

	// Some commands trigger a block read or write, so in those cases
	// we don't want to do the unselect just yet...
	switch(cmd) {
		case READ_SINGLE_BLK:
		case READ_MULTIPLE_BLK:
		case WRITE_BLK:
		case WRITE_MULTIPLE_BLK:
		case STOP_TRANSMISSION:
		case SEND_CSD:
		case SEND_CID:
			break;
		default:
			UNSELECT_SD();
			SPI_DISABLE();
			__builtin_bfin_ssync();
			break;
	}
		
	return(rc);
}

/* sdReadCxD():
 * Read the SDCard's ID (CID) or Card-Specific-Data (CSD) register...
 */
int
sdReadCxD(int interface, uchar *buf, int cmd)
{
	int i;
	uchar	rsptok, resp[8];

	// Retrieve the CID register:
	if (sdCardCmd(interface,cmd,0,resp) == -1)
		return(-1);

	while(1) {
		rsptok = sdSpiWrite(0xff);		
		if (rsptok == START_BLOCK_TOKEN)
			break;
		if ((rsptok & 0xf0) == 0x00) {
			if (sdVerbose & 2)
				printf("sdReadCxD: data_err_token: %02x\n",rsptok);
			return(-1);
		}
	}

	for(i=0;i<16;i++)
		buf[i] = sdSpiWrite(0xff);

	sdSpiWrite(0xff);			// CRC
	sdSpiWrite(0xff);			// CRC
	return(0);
}

/* sdInit():
 * This is an interface-specific wrapper for the call to sdGenericStartup().
 *
 * One snippet of info not in that spec is the which edge of the clock
 * the data is latched on (i.e. which mode of SPI).  Based on section
 * 3.9 "Bus Timing" of the microSD card spec from PQI (???), DI (in to
 * SD card) is latched * on the rising edge of the clock and DO
 * (out of SD card) is latched on the falling edge of the clock.
 *
 */
int
sdInit(int interface, int verbose)
{
	int	rc;

	if (interface != 0)
		return(-1);

	sdVerbose = verbose;

	sdio_init();

	*pSPI_BAUD = (SYSCLKFREQ/(SPISD_LO_SPEED*2));
	__builtin_bfin_ssync();

	SPI_ENABLE();
	__builtin_bfin_ssync();

	// Call sdGenericStartup() in the common code because this is
	// done the same way regardless of the underlying interface...
	if ((rc = sdGenericStartup(interface)) >= 0) {
		// Now that we're all initialized, we bump up the SPI
		// clock speed...
		*pSPI_BAUD = (SYSCLKFREQ/(SPISD_HI_SPEED*2));
		__builtin_bfin_ssync();
	}

	SPI_DISABLE();
	__builtin_bfin_ssync();
	return(rc);
}

int
waitforComplete(int tag)
{
    struct elapsed_tmr tmr;

	// Wait (with timeout) for completion:
    startElapsedTimer(&tmr,1000);
	while(1) {
		if (sdSpiWrite(0xff) == 0xff)
			break;
    	if(msecElapsed(&tmr)) {
			printf("WFC_%d timeout\n",tag);
			return(-1);
		}
	}
	return(0);
}

/* waitforDRT():
 * Wait for data-response-token.
 * Every data block written to the card is acknowledged with a data
 * response token (7.3.3.1 of spec).
 */
int
waitforDRT(int tag)
{
	uchar rsptok;
    struct elapsed_tmr tmr;

    startElapsedTimer(&tmr,1000);
	while(1) {
		rsptok = sdSpiWrite(0xff);		

		// The spec syas that after every block written to the card
		// it will respond with a data-response-token; however,
		// the only thing that seems to work for me is to wait for
		// 0xff and return success...
		if (rsptok == 0xff)
			return(0);

		if ((rsptok & DRT_FIXEDMASK) == DRT_FIXED) {
			switch(rsptok & DRT_STATMASK) {
				case DRT_STATACCEPTED:
					return(0);
				case DRT_STATCRCERR:
					printf("DRT%d crc err\n",tag);
					return(-1);
				case DRT_STATWRITEERR:
					printf("DRT%d write err\n",tag);
					return(-1);
				default:
					printf("DRT%d ??? err\n",tag);
					break;
			}
		}
    	if(msecElapsed(&tmr)) {
			printf("waitforDRT_%d timeout\n",tag);
			break;
		}
	}
	return(-1);
}

int
sdRead(int interface, char *buf, int blk, int blkcnt)
{
	int i;
	ulong addr;
	uchar resp[8];

	if (interface != 0)
		return(-1);
	
	if (sdInfoTbl[interface].initialized == 0) {
		printf("sdRead interface uninitialized\n",interface);
		return(-1);
	}

	// addr: argument passed in with READ_SINGLE_BLK...
	// For standard capacity, the unit is 'byte'.
	// For high capacity sd card, the unit is 'block' (512).
	// As of this writing, I don't know how to tell the
	// difference between cards.
	if (sdInfoTbl[interface].highcapacity == 1)
		addr = blk;
	else
		addr = blk*SD_BLKSIZE;

	if (blkcnt == 1) {
		if (sdCardCmd(interface, READ_SINGLE_BLK, addr, resp) == -1)
			return(-1);

		while(1) {
			uchar rsptok = sdSpiWrite(0xff);		
			if (rsptok == START_BLOCK_TOKEN)
				break;
			if ((rsptok & 0xf0) == 0x00) {
				printf("Data_err_token1: %02x\n",rsptok);
				return(-1);
			}
		}

		for(i=0;i<SD_BLKSIZE;i++)
			*buf++ = sdSpiWrite(0xff);
		sdSpiWrite(0xff);			// CRC
		sdSpiWrite(0xff);			// CRC
	}
	else {
		if (sdCardCmd(interface, READ_MULTIPLE_BLK, addr, resp) == -1) {
			printf("READ_MULTIPLE_BLK cmd failed\n");
			return(-1);
		}

		while(blkcnt) {
		for(i=0;i<100000;i++) {
			uchar rsptok = sdSpiWrite(0xff);		
			if (rsptok == START_BLOCK_TOKEN)
				break;
			if ((rsptok & 0xf0) == 0x00) {
				printf("Data_err_token2: %02x\n",rsptok);
				return(-1);
			}
		}
		if (i == 100000) {
			printf("WFT timeout\n");
			return(-1);
		}

			for(i=0;i<SD_BLKSIZE;i++)
				*buf++ = sdSpiWrite(0xff);
			sdSpiWrite(0xff);			// CRC
			sdSpiWrite(0xff);			// CRC
			blkcnt--;
		}

		if (sdCardCmd(interface, STOP_TRANSMISSION, 0, resp) == -1) {
			printf("READ_STOP_TRANSMISSION cmd failed\n");
			return(-1);
		}
		waitforComplete(1);
	}

	UNSELECT_SD();
	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}



int
sdWrite(int interface, char *buf, int blk, int blkcnt)
{
	int i;
	ulong addr;
	ushort crc;
	uchar resp[8];

	if (interface != 0)
		return(-1);

	if (sdInfoTbl[interface].initialized == 0) {
		printf("sdWrite interface uninitialized\n",interface);
		return(-1);
	}

	// addr: see note in sdRead().
	if (sdInfoTbl[interface].highcapacity == 1)
		addr = blk;
	else
		addr = blk*SD_BLKSIZE;

	if (blkcnt == 1) {
		// Send command:
		if (sdCardCmd(interface, WRITE_BLK, addr, resp) == -1)
			return(-1);

		sdSpiWrite(START_BLOCK_TOKEN);		

		// Send block:
		crc = xcrc16((uchar *)buf,(ulong)SD_BLKSIZE);
		for(i=0;i<SD_BLKSIZE;i++)
			sdSpiWrite(*buf++);

		// Send CRC:
		sdSpiWrite((crc & 0xff00)>>8);
		sdSpiWrite((crc & 0x00ff));	

		waitforDRT(1);
	}
	else {
		if (sdCardCmd(interface, WRITE_MULTIPLE_BLK, addr, resp) == -1) {
			printf("WRITE_MULTIPLE_BLK cmd failed\n");
			return(-1);
		}

		while(blkcnt) {
			sdSpiWrite(START_BLKMBW_TOKEN);		

			// Send block:
			crc = xcrc16((uchar *)buf,(ulong)SD_BLKSIZE);
			for(i=0;i<SD_BLKSIZE;i++)
				sdSpiWrite(*buf++);

			// Send CRC:
			sdSpiWrite((crc & 0xff00)>>8);
			sdSpiWrite((crc & 0x00ff));	

			// Wait for data response token:
			waitforDRT(2);

			blkcnt--;
		}

		// For multiple-block write, we end the transmission with
		// the STOP_TRAN_TOKEN (not the STOP_TRANSMISSION command)
		// (refer to 7.3.3.2 of SPLS2.0)
		sdSpiWrite(STOP_TRAN_TOKEN);		

		waitforComplete(2);
	}

	UNSELECT_SD();
	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(0);
}

void
sdio_init(void)
{
	*pPORTF_FER |= 0x3800;
	*pPORTFIO_DIR |= PF5;
	*pPORTFIO_SET |= PF5;

	*pSPI_FLG |= SD_FLS;
	__builtin_bfin_ssync();
}

/* sdInstalled():
 * Return 1 if card is installed, 0 if card is not installed
 * and -1 if the hardware cannot detect the state of the card.
 */
int
sdInstalled(int interface)
{
	return(-1);
}


#endif
