/* mode3boot.c:
 *
 * The purpose of this image is to be loaded (through SPI) by the ROM
 * loader in the Blackfin and upon load/run, do some basic memory test
 * of SDRAM and FLASH, then (assuming success of the tests) transfer
 * uMon to RAM and jump into it.
 */
#define __ADSPLPBLACKFIN__

#define SPIFBAUD 25000000

#include "config.h"
#include "cpu.h"
#include "cpuio.h"
#include "cdefBF537.h"
#include "spi_bf537.h"
#include "spif.h"
#include "spifmap.h"

#define BLINK_DELAY	5000000

#define _FAILURE_TEST_			0x2
#define _FAILURE_BAD_SPI_ID_	0x3
#define _FAILURE_BAD_DRAM_		0x4
#define _FAILURE_BAD_FLASH_		0x5
#define _FAILURE_SPI_WFR_		0x6
#define _FAILURE_SPI_WAIT_		0x7
#define _FAILURE_SPI_TXSWAIT_	0x8
#define _FAILURE_SPI_RXSWAIT_	0x9

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned long ulong;

unsigned long MonStack[32];

#define SPI_ENABLE()	*pSPI_CTL |= SPE
#define SPI_DISABLE()	*pSPI_CTL &= ~SPE

#define SPI_WRITE_MODE()	\
	*pSPI_CTL &= ~TDBR_CORE	// RDBR-read initiates transaction

#define SPI_READ_MODE()		\
	*pSPI_CTL |= TDBR_CORE	// TDBR-write initiates transaction

/* target_putchar():
 * When buffer has space available, load the incoming character
 * into the UART.
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit */
	while(!(*pUART0_LSR & THRE));

	*pUART0_THR = (unsigned short)c;

    return((int)c);
}

int
target_gotachar(void)
{
	if (*pUART0_LSR & DR)
		return(1);
    return(0);
}
 
int 
target_getchar(void)
{
    return((int)*pUART0_RBR);
}

int
getchar(void)
{
	while(!target_gotachar());
    return((int)target_getchar());
}


void
pstring(char *str)
{
	while(*str)
		target_putchar(*str++);
}

void
prhex(long lval)
{
	int idx, shift;
	unsigned long mask = 0xf0000000;

	target_putchar('0');
	target_putchar('x');
	shift = 28;
	while(mask) {
		idx = (lval & mask) >> shift;
		if (idx < 10)
			target_putchar(idx | 0x30);
		else
			target_putchar(idx + 55);
		mask >>= 4;
		shift -= 4;
	}
	return;
}

/* failure():
 * Endless loop that blinks the RED led with the specified value
 */
void
failure(int num)
{
	int i, j;

	DS2_OFF();
	DS3_OFF();

	while(1) {
		for(i=0;i<num;i++) {
			DS2_RED();
			for(j=0;j<BLINK_DELAY;j++);
			DS2_OFF();
			for(j=0;j<BLINK_DELAY;j++);
		}
		for(j=0;j<(BLINK_DELAY*10);j++);
		if (target_gotachar()) {
			target_getchar();
			target_putchar('E');
			prhex(num);
			target_putchar('\r');
			target_putchar('\n');
		}
	}
}

void
ledtoggle(void)
{
	volatile int i;

	DS2_OFF();
	DS3_OFF();

	DS2_ON();
	DS3_ON();
	for(i=0;i<BLINK_DELAY;i++);
	DS2_OFF();
	DS3_OFF();
	for(i=0;i<BLINK_DELAY;i++);
}

/* devInit():
 * Initialize console...
 */
void
devInit(int baud)
{
	unsigned short tmp;
	unsigned short	lctl, hi, lo;

	*pUART0_GCTL = UCEN;		/* Enable UART clocks */
	*pUART0_MCR = 0;
	*pUART0_LCR = 0x03;
	tmp = *pUART0_LSR;
	*pUART0_IER = 0;
	*pUART0_SCR = 0;

	lo = SYSCLKFREQ/(baud*16);
	hi = (lo & 0xff00) >> 8;
	lo &= 0x00ff;

	lctl = *pUART0_LCR;
	*pUART0_LCR = DLAB;
	*pUART0_DLL = lo;
	*pUART0_DLH = hi;
	*pUART0_LCR = lctl;
}

int
spifwait(void)
{
	volatile int i;
	volatile unsigned short stat;

	__builtin_bfin_ssync();
	for(i=0;i<200000;i++) {
		stat = *pSPI_STAT;
		__builtin_bfin_ssync();
		if (stat & SPIF)
			break;
	} 
	if (i == 200000) {
		SPI_DISABLE();
		__builtin_bfin_ssync();
		failure(_FAILURE_SPI_WAIT_);
		return(-1);
	}

	__builtin_bfin_ssync();
	return(0);
}

/* spitxswait():
 * Wait for transmit buffer empty.
 */
int
spitxswait(void)
{
	volatile int i;
	volatile unsigned short stat;

	__builtin_bfin_ssync();
	for(i=0;i<200000;i++) {
		stat = *pSPI_STAT;
		__builtin_bfin_ssync();
		if ((stat & TXS) == 0)
			break;
	} 
	if (i == 200000) {
		SPI_DISABLE();
		__builtin_bfin_ssync();
		failure(_FAILURE_SPI_TXSWAIT_);
		return(-1);
	}

	__builtin_bfin_ssync();
	return(0);
}

/* spirxswait():
 * Wait for receive buffer full.
 */
int
spirxswait(void)
{
	volatile int i;
	volatile unsigned short stat;

	__builtin_bfin_ssync();
	for(i=0;i<200000;i++) {
		stat = *pSPI_STAT;
		__builtin_bfin_ssync();
		if (stat & RXS)
			break;
	} 
	if (i == 200000) {
		SPI_DISABLE();
		__builtin_bfin_ssync();
		failure(_FAILURE_SPI_RXSWAIT_);
		return(-1);
	}

	__builtin_bfin_ssync();
	return(0);
}

int
spifInit(void)
{
	/* The AT25DF321 is on SPI CS1 line, so enable that line.
	 */
	*pSPI_FLG |= FLS1;
	__builtin_bfin_ssync();

	*pSPI_BAUD = (SYSCLKFREQ/(SPIFBAUD*2));
	__builtin_bfin_ssync();

	/* Enable SPI interface as...
	 *  Master, MSB first, 
	 */
	*pSPI_CTL = (SPE | MSTR | CPOL | CPHA);
	__builtin_bfin_ssync();

	*pSPI_STAT = (MODF | TXE | RBSY | TXCOL);
	__builtin_bfin_ssync();

	return(0);
}

int
spifWaitForReady(void)
{
	unsigned short stat;
	volatile int timeout = 1000000;

	while(timeout > 0) {
		stat = spifReadStatus(0);	
		if ((stat & BSY) == 0)
			break;
		timeout--;
	}
	if (timeout == 0) {
		failure(_FAILURE_SPI_WFR_);
		return(-1);
	}
	return(0);
}

int
spifId(int notused)
{
	int i, rc;
	volatile unsigned short tmp;
	volatile unsigned char buf[3];

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	/* Write id request command:
	 */
	*pSPI_TDBR = 0x009f;	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE */
	spifwait();

	SPI_READ_MODE();
	tmp = *pSPI_RDBR;
	spifwait();

	for(i=0;i<3;i++) {			/* Retrieve 3 bytes from device. */
		*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
		spirxswait();			/* Wait for receive buffer full */
		buf[i] = (unsigned char)*pSPI_RDBR;
	}

	spifwait();
	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	/* Make sure the devid is 0x1f4700...
	 * (don't check the lsb)
	 */
	rc = 0;
	if ((buf[0] != 0x1f) || (buf[1] != 0x47))
		failure(_FAILURE_BAD_SPI_ID_);

	return(rc);
}

unsigned short
spifReadStatus(int notused)
{
	volatile unsigned short tmp;
	volatile unsigned short stat;

	SPI_WRITE_MODE();

	*pSPI_FLG &= FLG1;

	/* Write id request command:
	 */
	*pSPI_TDBR = 0x0005;	
	tmp = *pSPI_RDBR;		/* Dummy read initiates the WRITE */
	spifwait();

	SPI_READ_MODE();
	tmp = *pSPI_RDBR;
	spifwait();

	*pSPI_TDBR = 0;			/* Dummy write initiates the READ. */
	spirxswait();			/* Wait for receive buffer full */
	stat = *pSPI_RDBR;

	spifwait();
	*pSPI_FLG |= ~FLG1;

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(stat);
}

int
spifReadBlock(unsigned long addr,char *data,int len)
{
	int i;
	volatile unsigned short tmp;
	unsigned char buf[5];

	SPI_ENABLE();
	__builtin_bfin_ssync();

	SPI_WRITE_MODE();
	__builtin_bfin_ssync();

	*pSPI_FLG &= FLG1;
	__builtin_bfin_ssync();

	buf[0] = 0x0b;
	buf[1] = ((addr & 0xff0000) >> 16);
	buf[2] = ((addr & 0xff00) >> 8);
	buf[3] = ((addr & 0xff));
	buf[4] = 0x00;				// Don't care

	for(i=0;i<5;i++) {
		*pSPI_TDBR = buf[i];	// Load xmit register
		__builtin_bfin_ssync();
		tmp = *pSPI_RDBR;		// Do dummy read to kick off the transfer.
		if (spifwait() != 0)
			return(-1);
	}

	SPI_READ_MODE();
	tmp = *pSPI_RDBR;
	spifwait();

	for(i=0;i<len;i++) {	
		*pSPI_TDBR = 0;	
		spirxswait();
		data[i] = *pSPI_RDBR;
	}

	spifwait();

	*pSPI_FLG |= ~FLG1;
	__builtin_bfin_ssync();

	SPI_DISABLE();
	__builtin_bfin_ssync();

	return(len);
}

void
memtest(void)
{
	volatile long i, *lp, errtot;

	errtot = 0;
	lp = (volatile long *)BRAMBASE;

	for(i=0;i<(1024*256);i++) 
		*lp++ = i;

	lp = (volatile long *)BRAMBASE;
	for(i=0;i<(1024*256);i++) {
		if (*lp != i)
			failure(_FAILURE_BAD_DRAM_);
		lp++;
	}
}

#define FLASHID 0x2022cb

/* flashtest():
 * Query the flash for its device ID.  If that doesn't succeed, assume
 * failure.
 */
void
flashtest(void)
{
#if INCLUDE_FLASH
	ulong	id;
	ushort	man, dev, val;

	/* Issue "auto-select" command and retrieve manufacturer and
	 * device ID.
	 */
	*(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR+(0x555<<1)) = 0x00aa;
	asm("csync;");
	*(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR+(0x2aa<<1)) = 0x0055;
	asm("csync;");
	*(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR+(0x555<<1)) = 0x0090;
	asm("csync;");

	man = *(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR);
	asm("csync;");
	dev = *(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR + 2);
	asm("csync;");

	/* Issue read-reset command to go back to read-array mode.
	 */
	*(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR) = 0x00f0;
	val = *(volatile unsigned short *)(FLASH_BANK0_BASE_ADDR);

	id = man;
	id <<= 16;
	id |= dev;

	if (id != FLASHID) 
		failure(_FAILURE_BAD_FLASH_);
#endif
}

void
mode3start(void)
{
	int xdone, i;
	volatile long *lp;
	void (*entry)(void);
	volatile char *cp;
	volatile int c;
	volatile unsigned short stmp;
	volatile unsigned long ltmp;

	/* PLL setup:
	 */
	asm("ssync;");
	*pPLL_CTL = (MSEL_VAL << 9) | PLL_DF_VAL;
	*pSIC_IWR = 1;
	asm("ssync;");
	asm("cli r0;");
	asm("idle;");
	asm("sti r0;");
	asm("ssync;");

	*pPLL_LOCKCNT =	0x0300;
	*pPLL_DIV = (SSEL_VAL) | (CSEL_VAL << 4);
	asm("ssync;");

	*pDTEST_COMMAND = 0;

	/* Memory bank init:
	 */
	*pEBIU_AMBCTL0 = 0x7bb07bb0;
	*pEBIU_AMBCTL1 = 0xffc27bb0;
	asm("ssync;");

	/* SDRAM ctrlr init:
	 * (should be done 100us after reset)
	 * Refer to pg 6-60 of HW reference.
	 */
	for(c=0;c<10000000;c++);
	*pEBIU_SDRRC = (((SYSCLKFREQ/1000) * 64) / 8192) - (TRAS + TRP);
	asm("ssync;");
	*pEBIU_SDBCTL = (EBCAW_10 | EBSZ_64 | EBE);
	asm("ssync;");
	if (*pEBIU_SDSTAT & SDRS)
		*pEBIU_SDGCTL = (PSS | TWR_2 | TRCD_2 | TRP_2 | TRAS_5 | CL_3 | SCTLE);
	else
		*pEBIU_SDGCTL = (TWR_2 | TRCD_2 | TRP_2 | TRAS_5 | CL_3 | SCTLE);
	asm("ssync;");

	*pEBIU_AMGCTL = 0x00ff;
	asm("ssync;");
	
	for(c=0;c<10000;c++);
		*(char *)c = (char)c;

	*pPORT_MUX = 0x0000;

#if HARDWARE_IS_REVA

	/* Port F configuration:
	 */
	stmp = *pPORTF_FER;
	*pPORTF_FER = 0x870f;	/* Set the bits that are NOT used as GPIO */
	*pPORTF_FER = 0x870f;
	*pPORTFIO_DIR = LED_BITMASK | ISO_OUT1BIT | ISO_OUT2BIT | TX485_EN;
	*pPORTFIO_CLEAR = TX485_EN | ISO_OUT1BIT | ISO_OUT2BIT;
	*pPORTFIO_INEN = ISO_INBIT;

	/* Port G configuration:
	 */
	stmp = *pPORTG_FER;
	*pPORTG_FER = 0x00ff;
	*pPORTG_FER = 0x00ff;
	*pPORTGIO_DIR =
		CAMERA_RESET_BIT | DECODER_RESET_BIT | EXPOSURE_BIT | AB_SELECT_BIT;

#elif HARDWARE_IS_REVB

	/* Port F configuration:
	 */
	stmp = *pPORTF_FER;
	*pPORTF_FER = 0xbf0f;	/* Set the bits that are NOT used as GPIO */
	*pPORTF_FER = 0xbf0f;
	*pPORTFIO_DIR = LED_BITMASK | ISO_OUT2BIT | TX485_EN;
	*pPORTFIO_CLEAR = TX485_EN | ISO_OUT2BIT;

	/* Port G configuration:
	 */
	stmp = *pPORTG_FER;
	*pPORTG_FER = 0x00ff; 	
	*pPORTG_FER = 0x00ff;
	*pPORTGIO_DIR = CAMERA_RESET_BIT | ISO_OUT1BIT | EXPOSURE_BIT;
	*pPORTGIO_INEN = ISO_INBIT;

#endif

	/* Port H configuration:
	 */
	stmp = *pPORTH_FER;
	*pPORTH_FER = 0xffff;
	*pPORTH_FER = 0xffff;

	DS2_OFF();
	DS3_OFF();

	devInit(57600);
	memtest();
	flashtest();
	spifInit();
	spifId(0);

	/* Memory map of the 4Mbyte of SPI flash (areas not shown in
	 * proportion):
	 * 
	 *  0x3fffff -------------
	 *           |           |
	 *           |           |
	 *           | TFS SPACE |
	 *           |           |
	 *           |           |
	 *           |           |
	 *           |           |
	 *           |           |
	 *  0x040000 -------------
	 *           |           |
	 *           | bram.bin  |
	 *           |  (uMON)   |
	 *           |           |
	 *  0x004000 -------------
	 *           | mode3.bin |
	 *  0x000000 -------------
	 * 
	 */ 
	/* Transfer bram.bin (uMon) image from SPI flash to runtime
	 * location...
	 */
	spifReadBlock(SPIF_UMON_IMAGE_BASE,(char *)BRAMBASE,SPIF_UMON_IMAGE_SIZE);	

#if ((FLASHRAM_END-FLASHRAM_BASE+1) != SPIF_TFS_SIZE)
#error: TFS in SPI is different size than TFS in RAM
#endif

	/* Clean up the area to be used by TFS-RAM by writing 0xffs
	 * to that space...
	 */
	lp = (volatile long *)FLASHRAM_BASE;
	while(lp < (long *)FLASHRAM_END)
		*lp++ = 0xffffffff;

	/* Transfer the next 3Mg to 0x3800000 in DRAM.
	 */
	spifReadBlock(SPIF_TFS_BASE,(char *)FLASHRAM_BASE,SPIF_TFS_SIZE);	

	/* Turn on all LEDs, and jump into uMon base address...
	 */
	DS2_ON();
	DS3_ON();

	entry = (void(*)(void))BRAMBASE;
	entry();

	/* Should not get here. */
	while(1) {
		if (target_gotachar()) {
			c = target_getchar();
			if (c == 'e')
				failure(_FAILURE_TEST_);
			else if (c == 'y')
				target_putchar('Y');
			else
				target_putchar(c);
			
		}
		ledtoggle();
	}
}
