//=============================================================================
//
//      pci_405.h
//
//      PCI Support
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         2002-04-06
// Description:  This file contains the PCI defines and access macros 
//
//
//=============================================================================

//-----------------------------------------------------------------------------
// Base address of PCI Registers
#define PCI_405(_x_)  			*(vulong *)(0xef400000 + _x_)
#define PCI_CFG_ADD  			*(vulong *)(0xeec00000)
#define PCI_CFG_DATA_32(_x_)  	*(vulong *)(0xeec00004 | (_x_ & 0x3))
#define PCI_CFG_DATA_16(_x_)  	*(vushort *)(0xeec00004 |(_x_ & 0x3))
#define PCI_CFG_DATA_8(_x_)  	*(vuchar *)(0xeec00004 | (_x_ & 0x3))
#define PCI_CFG_EN				0x80000000

// GPIO Bit
#define GPIO0_OR				*(vulong *)(0xef600700)
#define GPIO0_TS				*(vulong *)(0xef600704)
#define IDSEL1					0x00010000
#define IDSEL2					0x00008000
#define IDSEL					((GPIO0_OR & 0x18000) >> 15)

#define DEV_START				11
//-----------------------------------------------------------------------------
// Mappings for PCI memory and IO spaces
//

// The 405GP/r implements 3 Local bus to PCI Apertures for Memory Space, an
// aperture for Local to PCI I/O Space (0 to 64KB PCI I/O Addresses) and another
// aperture for Lcoal to PCI I/O Space (8MB to 4GB PCI I/O Addresses).
// Configuration is handled via a Config Address Register and a Config Data
// Register.  IDSEL is handled via GPIO bits 15 (SLOT 2) and 16 (SLOT 1)
//
//
// These are the addresses, in CPU space, where the PCI apertures are
// for each type of PCI access.  This is also where the PCI devices
// will be mapped when found for simplicity.
#define PCI_MEM_BASE 		0xA0000000		// programmed into PCIL0_PMM0PCILA and PMM0PCIHA
#define PCI_MEM_SIZE		0x10000000		// 256Mbytes

#define PCI_IO_BASE 		0xE8800000		// fixed by 405GP Bridge
#define PCI_IO_SIZE			0x02000000		// 32Mbytes

#define PCI_SM_BASE			0x01000000		// 1Mbyte	shared memory space
#define PCI_SM_SIZE			0x01000000

//--------------------------------------------------------------------------
// Host to Little Endian Macros
// Big Endian Host
#define HTOLE16(x) ((((x) & 0xff00U) >> 8) | (((x) & 0x00ffU) << 8))
#define HTOLE32(x) ((((x) & 0xff000000UL) >> 24) | (((x) & 0x00ff0000UL) >> 8 ) | (((x) & 0x0000ff00UL) << 8 ) | (((x) & 0x000000ffUL) << 24))

// Little Endian Host
//#define HTOLE16(x) (x)
//#define HTOLE32(x) (x)

//-----------------------------------------------------------------------------
// Macros for reading and writing the shared memory structures
//-----------------------------------------------------------------------------
#define RD_SM8(_reg_,_val_)  ((_val_) = *((vuchar *)((((ulong)_reg_) ^ (0)))))
#define WR_SM8(_reg_,_val_)  (*((vuchar *)((((ulong)_reg_) ^ (0)))) = (_val_))

#define RD_SM16(_reg_,_val_) ((_val_) = HTOLE16(*((vushort *)((((ulong)_reg_) ^ (0))))))
#define WR_SM16(_reg_,_val_) (*((vushort *)((((ulong)_reg_) ^ (0)))) = HTOLE16(_val_))

#define RD_SM32(_reg_,_val_) ((_val_) = HTOLE32(*((vulong *)((ulong)_reg_))))
#define WR_SM32(_reg_,_val_) (*((vulong *)((ulong)_reg_)) = HTOLE32(_val_))

//-----------------------------------------------------------------------------
// 405 PCI Configuration Registers accessed from CPU Space
//
#define PCIL0_PMM0LA			PCI_405(0x00) 					// PMM 0 Local Address
#define PCIL0_PMM0MA 			PCI_405(0x04) 					// PMM 0 Mask/Attribute
#define PCIL0_PMM0PCILA 		PCI_405(0x08) 					// PMM 0 PCI Low Address
#define PCIL0_PMM0PCIHA 		PCI_405(0x0C) 					// PMM 0 PCI High Address
#define PCIL0_PMM1LA 			PCI_405(0x10) 					// PMM 1 Local Address
#define PCIL0_PMM1MA 			PCI_405(0x14) 					// PMM 1 Mask/Attribute
#define PCIL0_PMM1PCILA 		PCI_405(0x18) 					// PMM 1 PCI Low Address
#define PCIL0_PMM1PCIHA 		PCI_405(0x1C) 					// PMM 1 PCI High Address
#define PCIL0_PMM2LA 			PCI_405(0x20) 					// PMM 2 Local Address
#define PCIL0_PMM2MA 			PCI_405(0x24) 					// PMM 2 Mask/Attribute
#define PCIL0_PMM2PCILA 		PCI_405(0x28) 					// PMM 2 PCI Low Address
#define PCIL0_PMM2PCIHA 		PCI_405(0x2C) 					// PMM 2 PCI High Address
#define PCIL0_PTM1MS 			PCI_405(0x30) 					// PTM 1 Memory Size/Attribute
#define PCIL0_PTM1LA 			PCI_405(0x34) 					// PTM 1 Local Address
#define PCIL0_PTM2MS			PCI_405(0x38) 					// PTM 2 Memory Size/Attribute
#define PCIL0_PTM2LA			PCI_405(0x3C) 					// PTM 2 Local Address

//-----------------------------------------------------------------------------
// Bit Defines for 405GP PCI Registers accessed from CPU Space
//
// PCIL0_PMM0MA PMM 0 and 1 Mask/Attribute bit defines
//
#define PCIL0_PMM0MA_MASK			0xFFFFF000	// Address mask bits
#define PCIL0_PMM0MA_PRE			0x00000002 	// Read Prefetching Enable
#define PCIL0_PMM0MA_ENA			0x00000001	// PLB to PCI Memory Mapping Enable
#define PCIL0_PMM1MA_MASK			0xFFFFF000	// Address mask bits
#define PCIL0_PMM1MA_PRE			0x00000002 	// Read Prefetching Enable
#define PCIL0_PMM1MA_ENA			0x00000001	// PLB to PCI Memory Mapping Enable

// PCIL0_PTM1MS PTM 1 and 2 Mask/Attribute bit defines
// Actual addresses are set in PCIC0_PTM1BAR and PCIC0_PTM2BAR.
// These registers here define the masking/size of the window
// and the remap address to the locla bus
#define PCIL0_PTM1MS_MASK			0xFFFFF000	// Address mask bits
#define PCIL0_PTM1MS_ENA			0x00000001	// PLB to PCI Memory Mapping Enable
#define PCIL0_PTM2MS_MASK			0xFFFFF000	// Address mask bits
#define PCIL0_PTM2MS_ENA			0x00000001	// PLB to PCI Memory Mapping Enable
//
// 405GP Configuration Cycle Address Register bits and macros
#define PCIC0_CFGADDR_EN 			0x80000000	// Enable			
#define PCIC0_CFGADDR_BUS(_x_)		(((_x_) << 16) & 0x00FF0000)	// Bus Number 	
#define PCIC0_CFGADDR_DEV(_x_) 		(((_x_) << 11) & 0x0000F800)	// Device Number	
#define PCIC0_CFGADDR_FUNC(_x_) 	(((_x_) << 8) & 0x00000700) 	// Function Number	
#define PCIC0_CFGADDR_REG(_x_) 		(((_x_)) & 0x000000FC)		// Register Number	

//-----------------------------------------------------------------------------
// 405GP/r PCI Configuration Registers accessed from PCI Space.
// Use config bus 0 and device 0 to access.	 
//
// Common PCI Registers	- Use the standard PCI names defined in
// pci.h instead of these.  They are shown here FYI.
#define PCIC0_VENDID 				0x00 		// PCI Vendor ID
#define PCIC0_DEVID 				0x02 		// PCI Device ID
#define PCIC0_CMD 					0x04 		// PCI Command Register
#define PCIC0_STATUS 				0x06 		// PCI Status Register
#define PCIC0_REVID 				0x08 		// PCI Revision ID
#define PCIC0_PCICLS				0x09 		// PCI Class Register
#define PCIC0_CACHELS 				0x0C 		// PCI Cache Line Size
#define PCIC0_LATTIM				0x0D 		// PCI Latency Timer
#define PCIC0_HDTYPE 				0x0E 		// PCI Header Type
#define PCIC0_BIST 					0x0F 		// PCI Built In Self Test Control
#define PCIC0_PTM1BAR 				0x14 		// PCI PTM 1 BAR
#define PCIC0_PTM2BAR 				0x18 		// PCI PTM 2 BAR
#define PCIC0_SBSYSVID 				0x2C 		// PCI Subsystem Vendor ID
#define PCIC0_SBSYSID 				0x2E 		// PCI Subsystem ID
#define PCIC0_CAP 					0x34 		// PCI Capabilities Pointer
#define PCIC0_INTLN 				0x3C 		// PCI Interrupt Line
#define PCIC0_INTPN 				0x3D 		// PCI Interrupt Pin
#define PCIC0_MINGNT 				0x3E 		// PCI Minimum Grant
#define PCIC0_MAXLTNCY 				0x3F 		// PCI Maximum Latency
#define PCIC0_PCIICS 				0x44 		// PCI Interrupt Control/Status
//
// Extended PCI Registers for Bridge Control
#define PCIC0_ERREN 				0x48 		// Error Enable
#define PCIC0_ERRSTS 				0x49 		// Error Status
#define PCIC0_BRDGOPT1 				0x4A 		// PCI Bridge Options 1
#define PCIC0_PLBBESR0 				0x4C 		// PLB Slave Error Syndrome 0
#define PCIC0_PLBBESR1 				0x50 		// PLB Slave Error Syndrome 1
#define PCIC0_PLBBEAR 				0x54 		// PLB Slave Error Address Register
#define PCIC0_CAPID 				0x58 		// Capability Identifier
#define PCIC0_NEXTIPTR 				0x59 		// Next Item Pointer
#define PCIC0_PMC 					0x5A 		// Power Management Capabilities
#define PCIC0_PMCSR 				0x5C 		// Power Management Control Status
#define PCIC0_PMCSRBSE 				0x5E 		// PMCSR PCI-to-PCI Bridge Support Extensions
#define PCIC0_DATA 					0x5F 		// Data
#define PCIC0_BRDGOPT2 				0x60 		// PCI Bridge Options 2
#define PCIC0_PMSCRR 				0x64 		// Power Management State Change Request Register

// Bit defines for Extended PCI Registers
/*
#define PCI_CFG_CMD_MEM				0x02
#define PCI_CFG_CMD_MASTER			0x04
*/
//
// Error Enable	PCIC0_ERREN
#define PCIC0_ERREN_TAEE			0x40		// Target Abort Error Enable
#define PCIC0_ERREN_MERE_NONE 		0x00		// PLB Bus Error Response Enable = None
#define PCIC0_ERREN_MERE_SERR		0x10		// PLB Bus Error Response Enable = Assert PCISerr
#define PCIC0_ERREN_MERE_ABORT		0x20		// PLB Bus Error Response Enable = Abort Transaction
#define PCIC0_ERREN_MERE_ALL		0x30		// PLB Bus Error Response Enable = Assert PCISerr and Abort
#define PCIC0_ERREN_MEDE 			0x08		// PLB Master Error Detection Enable
#define PCIC0_ERREN_MEAE 			0x04		// PLB Bus Error Assertion Enable
#define PCIC0_ERREN_WDPE 			0x02		// Write Data Parity PCISerr Enable
#define PCIC0_ERREN_MAEE 			0x01		// Master Abort Error Enable

// Error Status	PCIC0_ERRSTS
#define PCIC0_ERRSTS_SARME 			0x10		// PCISErr Asserted on Received PLB Bus
#define PCIC0_ERRSTS_MED 			0x08		// PLB Bus Error Detected
#define PCIC0_ERRSTS_MEAE 			0x04		// PLB Bus Error Assertion Event
#define PCIC0_ERRSTS_WDPE 			0x02		// PCISerr on Write Data Parity Error
#define PCIC0_ERRSTS_PUR 			0x01		// PLB Unsupported Request

// PCI Bridge Options 1	- PCIC0_BRDGOPT1
#define PCIC0_BRDGOPT1_PMLTCR(_x_)	(((_x_) << 8) & 0xFF00) // PLB Master Latency Timer Count
#define PCIC0_BRDGOPT1_PLESE 		0x0080					// PLB Lock Error Status Enable
#define PCIC0_BRDGOPT1_PRP(_x_)		(((_x_) << 5) & 0x0060) // PLB Request Priority
#define PCIC0_BRDGOPT1_PGMAE 		0x0010					// PLB Guarded Memory Access Enable
#define PCIC0_BRDGOPT1_PAPM 		0x0008					// PCI Arbiter Park Mode
#define PCIC0_BRDGOPT1_PTMRCI_1 	0x0000					// PCI Target Memory Read Command = 1
#define PCIC0_BRDGOPT1_PTMRCI_LINE 	0x0002					// PCI Target Memory Read Command = Line
#define PCIC0_BRDGOPT1_PTMRCI_MULTI 0x0004					// PCI Target Memory Read Command = Multiple
#define PCIC0_BRDGOPT1_APLRM Atomic PLB Line Read Mode

// PCI Bridge Options 2	- PCIC0_BRDGOPT2
#define PCIC0_BRDGOPT2_EWPCI 		0x2000					// External Write to PCI Command Interrupt 		
#define PCIC0_BRDGOPT2_DPR 			0x1000					// Drive PCI Reset 		
#define PCIC0_BRDGOPT2_PSTLTD(_x_)	(((_x_) << 8) & 0x0F00)	// Subsequent Target Latency Timer Duration 		
#define PCIC0_BRDGOPT2_PDTD 		0x0004					// PCI Discard Timer Disable 		
#define PCIC0_BRDGOPT2_HCE 			0x0001					// Host Configuration Enable 		



