//=============================================================================
//
//      etherdev.c
//
//      Ethernet Abstraction Layer for Micro Monitor
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         07/23/2003
// Description:  This file contains the interface layer between Micro Monitor
//               and the Ethernet driver for the AT91RM9200 on the CSB337.
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "at91rm9200.h"
#include "at91rm9200_emac.h"

#if INCLUDE_ETHERNET

// Receive Buffer Descriptor Header
typedef struct
{
	ulong address;		// receive buffer address
	ulong status;		// status
} RXBUF_HDR;

// Bit assignments for Receive Buffer Descriptor
// Address - Word 0
#define RXBUF_ADD_BASE_MASK			0xfffffffc	// Base address of the receive buffer in memory
#define RXBUF_ADD_WRAP				BIT1		// If this bit is set, this is the last buffer in the ring
#define RXBUF_ADD_OWNED				BIT0		// 1 = SW owns the pointer, 0 = MAC owns buffer

// Status - Word 1
#define RXBUF_STAT_BCAST			BIT31		// Global all ones broadcast address detected
#define RXBUF_STAT_MULTI			BIT30		// Multicast hash match
#define RXBUF_STAT_UNI				BIT29		// Unicast hash match
#define RXBUF_STAT_EXT				BIT28		// External address (optional)
#define RXBUF_STAT_UNK				BIT27		// Unknown source address (reserved for future use)
#define RXBUF_STAT_LOC1				BIT26		// Local address match (Specific address 1 match)
#define RXBUF_STAT_LOC2				BIT25		// Local address match (Specific address 2 match)
#define RXBUF_STAT_LOC3				BIT24		// Local address match (Specific address 3 match)
#define RXBUF_STAT_LOC4				BIT23		// Local address match (Specific address 4 match)
#define RXBUF_STAT_LEN_MASK			0x3ff		// Length of frame including FCS

// Bit defines for the LXT971 Status Register #1 (phy address 0x01)
#define LXT971_SREG1_100BASE_T4		BIT15		// 1 = PHY able to perform 100BASE-T4
#define LXT971_SREG1_100BASE_X_FDX	BIT14		// 1 = PHY able to perform full-duplex 100BASE-X
#define LXT971_SREG1_100BASE_X_HDX	BIT13		// 1 = PHY able to perform half-duplex 100BASE-X
#define LXT971_SREG1_10BASE_FDX		BIT12		// 1 = PHY able to operate at 10 Mbps in full-duplex mode
#define LXT971_SREG1_10BASE_HDX		BIT11		// 1 = PHY able to operate at 10 Mbps in half-duplex mode
#define LXT971_SREG1_EXT_STATUS		BIT8		// 1 = Extended status information in register 15
#define LXT971_SREG1_MF_PREAMBLE	BIT6		// 1 = PHY will accept management frames with preamble suppressed
#define LXT971_SREG1_AUTO_NEG_DONE	BIT5		// 1 = Auto-negotiation complete
#define LXT971_SREG1_REM_FLT		BIT4		// 1 = Remote fault condition detected
#define LXT971_SREG1_AUTO_NEG_ABLE	BIT3		// 1 = PHY is able to perform Auto-Negotiation
#define LXT971_SREG1_LINK_UP		BIT2		// 1 = Link is up
#define LXT971_SREG1_JABBER			BIT1		// 1 = Jabber condition detected
#define LXT971_SREG1_EXT_REG		BIT0		// 1 = Extended register capabilities

// Bit defines for the LED configuration register (0x14)
#define LXT971_LED1(_x_)			((_x_ & 0xf) << 12)	// macros to set the led fields.  they use the same settings
#define LXT971_LED2(_x_)			((_x_ & 0xf) << 8)
#define LXT971_LED3(_x_)			((_x_ & 0xf) << 4)
#define LXT971_LED_SPEED			0x0 		// Display Speed Status (Continuous, Default)
#define LXT971_LED_TX				0x1 		// Display Transmit Status (Stretched)
#define LXT971_LED_RX				0x2 		// Display Receive Status (Stretched)
#define LXT971_LED_COL				0x3 		// Display Collision Status (Stretched)
#define LXT971_LED_LINK				0x4 		// Display Link Status (Continuous)
#define LXT971_LED_DX				0x5 		// Display Duplex Status (Continuous)
#define LXT971_LED_RX_TX			0x7 		// Display Receive or Transmit Activity (Stretched)
#define LXT971_LED_ON				0x8 		// Test mode- turn LED on (Continuous)
#define LXT971_LED_OFF				0x9 		// Test mode- turn LED off (Continuous)
#define LXT971_LED_BLINK_FAST		0xa 		// Test mode- blink LED fast (Continuous)
#define LXT971_LED_BLINK_SLOW		0xb 		// Test mode- blink LED slow (Continuous)
#define LXT971_LED_LINK_RX			0xc 		// Display Link and Receive Status combined 2 (Stretched)
#define LXT971_LED_LINK_RX_TX		0xd 		// Display Link and Activity Status combined 2 (Stretched)
#define LXT971_LED_DX_COL			0xe 		// Display Duplex and Collision Status combined 4 (Stretched)
#define LXT971_LED_STRETCH_30MS		(0 << 2)
#define LXT971_LED_STRETCH_60MS		(1 << 2)
#define LXT971_LED_STRETCH_100MS	(2 << 2)
#define LXT971_LED_STRETCH_ON		BIT1

// use internal SRAM for buffers and descriptors
// this also insures that the receive descriptors
// start on a 64byte boundary
//RXBUF_HDR *rxbuf_hdrs = (RXBUF_HDR *)(SRAM_BASE + 0x100);	// an array of receive buffer descriptors
RXBUF_HDR *rxbuf_hdrs = (RXBUF_HDR *)SRAM_BASE;	// an array of receive buffer descriptors
														// we leave room for ram based vectors
//uchar *txbuf = (uchar *)(SRAM_BASE + 0x200);		// one transmit buffer, 1536 bytes maximum
uchar *txbuf = (uchar *)(SRAM_BASE + 0x100);		// one transmit buffer, 1536 bytes maximum
uchar *rxbuf =  (uchar *)(SRAM_BASE + 0x800);		// receive buffers starting address

#define RX_BUFFER_SIZE 1600
#define NUM_RXBDS	   4

#define EMAC_DBG	   0

// ********************************************
// phyread():
// Read the PHY
ulong phyread(uchar reg)
{
	EMAC_REG(EMAC_MAN) = (0x01 << 30	// Start of Frame Delimiter
					    | 0x02 << 28	// Operation, 0x01 = Write, 0x02 = Read
						| 0x00 << 23	// Phy Number, 0 as we only have one
						| reg << 18		// Phy Register
						| 0x02 << 16);	// must be 0x02 for turn around field

	// wait for phy read to complete
	monDelay(100);

	if (EMAC_DBG) {
		printf("EMAC: Phy 0, Reg %d, Read 0x%04lx.\n", reg, (EMAC_REG(EMAC_MAN) & 0xffff));
	}

	return EMAC_REG(EMAC_MAN) & 0xffff;
}
// ********************************************
// phywrite()
// write the PHY
void phywrite(uchar reg, ushort data)
{
	EMAC_REG(EMAC_MAN) = (0x01 << 30		// Start of Frame Delimiter
	    			   | 0x01 << 28			// Operation, 0x01 = Write, 0x02 = Read
					   | 0x00 << 23			// Phy Number, 0 as we only have one
					   | reg << 18			// Phy Register
					   | 0x02 << 16			// must be 0x02 for turn around field
					   | data);
	if (EMAC_DBG) {
		printf("EMAC: Phy 0, Reg %d, Write 0x%04x.\n", reg, data);
	}

	// wait for phy write to complete
	monDelay(1);
}

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{

	// Disable TX/RX 
	EMAC_REG(EMAC_CTL)  &= ~(EMAC_CTL_TE | EMAC_CTL_RE);
}

/*
 * eninit():
 * This would include establishing buffer descriptor tables and
 * all the support code that will be used by the ethernet device.
 *
 * It can be assumed at this point that the array uchar BinEnetAddr[6]
 * contains the 6-byte MAC address.
 *
 * Return 0 if successful; else -1.
 */
int
eninit(void)
{
	int i, timeout;
	ulong emac_link_status = 0;

	if (EMAC_DBG) {
		printf("EMAC: RX Buffer Header Descriptor Address = 0x%08lx.\n", (ulong)rxbuf_hdrs);
	}

	// initialize our receive buffer descriptors
	for (i = 0; i < NUM_RXBDS-1; i++)
	{
		if (EMAC_DBG) {
			printf("EMAC: RX Buffer %d Address = 0x%08lx.\n", i, (ulong)(&rxbuf[i * RX_BUFFER_SIZE]));
		}
		rxbuf_hdrs[i].address = (ulong)(&rxbuf[i * RX_BUFFER_SIZE]);
		rxbuf_hdrs[i].status = 0x00000000;
	}
	// set the wrap bit on the last buffer
	if (EMAC_DBG) {
		printf("EMAC: RX Buffer %d (Last) Address = 0x%08lx.\n", i, (ulong)(&rxbuf[i * RX_BUFFER_SIZE]));
	}
	// last one needs the wrapbit set as well
	rxbuf_hdrs[i].address = (ulong)(&rxbuf[i * RX_BUFFER_SIZE]) | RXBUF_ADD_WRAP;
	rxbuf_hdrs[i].status = 0x00000000;
	
	// set the MAC address - must write the low first
	EMAC_REG(EMAC_SA1L) = (BinEnetAddr[2] << 24) | (BinEnetAddr[3] << 16) | (BinEnetAddr[4] << 8) | BinEnetAddr[5];
	EMAC_REG(EMAC_SA1H) = (BinEnetAddr[0] << 8) | BinEnetAddr[1];

	// point to our receive buffer queue
	EMAC_REG(EMAC_RBQP) = (ulong)rxbuf_hdrs;

	if (EMAC_DBG) {
		printf("EMAC: RX Buffer Queue = 0x%08lx.\n", (ulong)rxbuf_hdrs);
	}
	// clear any left over status bits
	EMAC_REG(EMAC_RSR)  &= ~(EMAC_RSR_OVR | EMAC_RSR_REC | EMAC_RSR_BNA);

	// set the MII clock divder to MCK/64 and set copy all frames
	EMAC_REG(EMAC_CFG) = (EMAC_CFG_CLK_64 | EMAC_CFG_CAF | EMAC_CFG_BIG | EMAC_CFG_FD);

	// enable the MII interface
	EMAC_REG(EMAC_CTL) = EMAC_CTL_MPE;

#if 1
	if (EMAC_DBG) {
		printf("EMAC: Getting Link Status.\n");
	}
	// Get the link status - wait for done with a timeout
	timeout = 20;
	while (timeout)
	{
		monDelay(2000);
		emac_link_status = phyread(0x01);
		if (!(emac_link_status & LXT971_SREG1_AUTO_NEG_ABLE)){
			printf("EMAC: PHY is unable to Auto-Negotatiate!\n");
			timeout = 0;
			break;
		}
		if (emac_link_status & LXT971_SREG1_AUTO_NEG_DONE)
		{
			if (EMAC_DBG) 
				printf("EMAC: Auto-Negotiate Complete, Link = ");
			break;
		}
		timeout-- ;
	}
	if (!(timeout))
	{
		printf("EMAC: Auto-Negotatiate Failed, Status = 0x%04lx!\n", emac_link_status);
		printf("EMAC: Initialization Halted.\n");
		return -1; 
	}

	// Set SPD and FD based on the return link status
	if (emac_link_status & (LXT971_SREG1_100BASE_X_FDX | LXT971_SREG1_100BASE_X_HDX))
	{
		EMAC_REG(EMAC_CFG) |= EMAC_CFG_SPD;
		if (EMAC_DBG) 
			printf("100MBIT, ");
	}
	else
	{
		EMAC_REG(EMAC_CFG) &= ~EMAC_CFG_SPD;
		if (EMAC_DBG) 
			printf("10MBIT, ");
	}		
			
	if (emac_link_status & (LXT971_SREG1_100BASE_X_FDX | LXT971_SREG1_10BASE_FDX))
	{
		EMAC_REG(EMAC_CFG) |= EMAC_CFG_FD;
		if (EMAC_DBG) 
			printf("Full Duplex.\n");
	}
	else
	{
		EMAC_REG(EMAC_CFG) &= ~EMAC_CFG_FD;
		if (EMAC_DBG) 
			printf("Half Duplex.\n");
	}		

#endif

	// Set PHY LED1 to Speed, PHY LED2 to combined Link/Activity and enable pulse stretching
	phywrite(0x14, (LXT971_LED1(LXT971_LED_SPEED)
				 | LXT971_LED2(LXT971_LED_LINK_RX_TX)
				 | LXT971_LED_STRETCH_30MS
				 | LXT971_LED_STRETCH_ON)); 

	if (EMAC_DBG) {
		printf("EMAC: Enable TX and RX.\n");
	}
	// Enable TX/RX and clear the statistics counters
	EMAC_REG(EMAC_CTL) = (EMAC_CTL_TE | EMAC_CTL_RE | EMAC_CTL_CSR);

	return (0);

}

int
EtherdevStartup(int verbose)
{
#if  CORE_AT_FULL_SPEED
	static int beenhere;
#endif

	/* Put ethernet controller in reset: */
	enreset();

#if  CORE_AT_FULL_SPEED
	/* This initial delay appears to be necessary to make auto-negotiation
	 * consistently succeed when CORE_AT_FULL_SPEED is enabled.
	 */
	if (beenhere == 0) {
		beenhere = 1;
		monDelay(1000);
	}
#endif


	/* Initialize controller: */
	eninit();

	return(0);
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * are any incoming ethernet packets.
 */
int
polletherdev(void)
{
	static int pktidx;
	int	pktlen = 0, pktcnt = 0;

#if 0
	ulong status32;

	// Clear any receive errors - we let the higher layer protocols handle
	// any issues
	status32 = EMAC_REG(EMAC_RSR);
	if (status32 & EMAC_RSR_BNA){
		printf("EMAC: RX Error Status - Buffer Not Available\n");
		EMAC_REG(EMAC_RSR) = EMAC_RSR_BNA;
		if (EMAC_DBG) {
			printf("EMAC: EMAC_CTL = 0x%08lx.\n", EMAC_REG(EMAC_CTL));
		}
	}

	if (status32 & EMAC_RSR_OVR){
		printf("EMAC: RX Error Status - Receiver Overrun\n");
		EMAC_REG(EMAC_RSR) = EMAC_RSR_OVR;
		if (EMAC_DBG) {
			printf("EMAC: EMAC_CTL = 0x%08lx.\n", EMAC_REG(EMAC_CTL));
		}
	}

#endif
	// clear the receive bit as we do not use it anyway
	EMAC_REG(EMAC_RSR) = (EMAC_RSR_REC | EMAC_RSR_OVR | EMAC_RSR_BNA);

	// See if the next buffer descriptor has a packet...
	if (rxbuf_hdrs[pktidx].address & RXBUF_ADD_OWNED)
	{
		int idx;

		idx = pktidx;
		pktidx++;
		if (pktidx == NUM_RXBDS)
			pktidx = 0;
		pktlen = rxbuf_hdrs[idx].status & RXBUF_STAT_LEN_MASK;
			
		if (EMAC_DBG) {
			printf("EMAC: RX Buffer %d Address = 0x%08lx, Status = 0x%08lx.\n", 
					idx, rxbuf_hdrs[idx].address, rxbuf_hdrs[idx].status);
		}
		pktcnt++;
		EtherRFRAMECnt++;
		processPACKET((struct ether_header *)(rxbuf_hdrs[idx].address & 0xfffffffc), pktlen);

		// Release the buffer back to the EMAC
		rxbuf_hdrs[idx].address &= ~RXBUF_ADD_OWNED;
	}

	// make sure the receiver/transmitter are still enabled
//	EMAC_REG(EMAC_CTL) |= (EMAC_CTL_RE | EMAC_CTL_TE);

	return(pktcnt);
}


/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.
 */
uchar *
getXmitBuffer(void)
{
	return txbuf;
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
	ulong status32;
	int timeout;

	if (EMAC_DBG) {
		printf("EMAC: TX Buffer Address = 0x%08lx, Length = %d.\n", (ulong)txbuf, length);
	}

#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)txbuf,length,ETHER_OUTGOING);
#endif

	/* Bump up the packet length to a minimum of 64 bytes.
	 */
	if (length < 64)
		length = 64;

	// tell the EMAC to send the tx buffer pointed to by txbuf
	// wait (with timeout) until the chip is ready
	timeout = 500000;
	while(timeout){
		if (EMAC_REG(EMAC_TSR) & EMAC_TSR_BNQ) break;
		timeout--;
	}
	if (timeout == 0)
	{
		printf("EMAC Timeout waiting for Ready to Transmit, Re-initializing EMAC!\n");
		enreset();
		eninit();
		return (0);
	}

	// clear any pending status
	status32 = EMAC_REG(EMAC_TSR);
	if (status32 & (EMAC_TSR_OVR | EMAC_TSR_COL | EMAC_TSR_RLE | EMAC_TSR_UND)){
		if (EMAC_DBG) {
			printf("EMAC: TX Error Status = 0x%08lx.\n", status32);
		}
	}
	EMAC_REG(EMAC_TSR) = (EMAC_TSR_OVR | EMAC_TSR_COL | EMAC_TSR_RLE
					   | EMAC_TSR_COMP | EMAC_TSR_UND); 

	// tell the EMAC about our buffer
	EMAC_REG(EMAC_TAR) = (ulong)txbuf;
	EMAC_REG(EMAC_TCR) = (ulong)length;
	EtherXFRAMECnt++;

	// wait until tx is complete.  This acts as a throttle to prevent
	// receiver overruns
	timeout = 10000;
	while(timeout){
		if (EMAC_REG(EMAC_TSR) & EMAC_TSR_COMP) break;
		timeout--;
	}

	return 0;
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */

void
ShowEtherdevStats()
{
	printf("EMAC Transmit Frames OK = %d\n",(int)EMAC_REG(EMAC_FRA));
	printf("EMAC Single Collision Frames = %d\n",(int)EMAC_REG(EMAC_SCOL));
	printf("EMAC Multiple Collision Frames = %d\n",(int)EMAC_REG(EMAC_MCOL));
	printf("EMAC Received Frames OK = %d\n",(int)EMAC_REG(EMAC_OK));
	printf("EMAC SEQE Errors = %d\n",(int)EMAC_REG(EMAC_SEQE));
	printf("EMAC Alignment Errors = %d\n",(int)EMAC_REG(EMAC_ALE));
	printf("EMAC Deffered Transmit Frames = %d\n",(int)EMAC_REG(EMAC_DTE));
	printf("EMAC Late Collision= %d\n",(int)EMAC_REG(EMAC_LCOL));
	printf("EMAC Excessive Collision Errors = %d\n",(int)EMAC_REG(EMAC_ECOL));
	printf("EMAC Carrier Sense Errors = %d\n",(int)EMAC_REG(EMAC_CSE));
	printf("EMAC Transmit Underrun Errors = %d\n",(int)EMAC_REG(EMAC_TUE));
	printf("EMAC Receive PHY RXER Errors = %d\n",(int)EMAC_REG(EMAC_CDE));
	printf("EMAC Receive Frame Too Big Errors = %d\n",(int)EMAC_REG(EMAC_ELR));
	printf("EMAC Receive Jabber Errors = %d\n",(int)EMAC_REG(EMAC_RJB));
	printf("EMAC Receive Frame Too Small Errors = %d\n",(int)EMAC_REG(EMAC_USF));
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */	 
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	return(1);
}

#endif
