/* reset.s:
 * First bit of boot code run by the processor.
 *
 * General notice:
 * This code is part of a boot-monitor package developed as a generic base
 * platform for embedded system designs.  As such, it is likely to be
 * distributed to various projects beyond the control of the original
 * author.  Please notify the author of any enhancements made or bugs found
 * so that all may benefit from the changes.  In addition, notification back
 * to the author will allow the new user to pick up changes that may have
 * been made by other users after this version of the code was distributed.
 *
 * Author:  Ed Sutter
 * email:   esutter@lucent.com 
 * phone:   908-582-2351
 *
 * This code was derived from source code from Motorola...
 *
 *********************************************************************
 *
 * Copyright:
 *  MOTOROLA, INC. All Rights Reserved.  
 *  You are hereby granted a copyright license to use, modify, and
 *  distribute the SOFTWARE so long as this entire notice is
 *  retained without alteration in any modified and/or redistributed
 *  versions, and that such modified versions are clearly identified
 *  as such. No licenses are granted by implication, estoppel or
 *  otherwise under any patents or trademarks of Motorola, Inc. This 
 *  software is provided on an "AS IS" basis and without warranty.
 *
 *  To the maximum extent permitted by applicable law, MOTOROLA 
 *  DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED, INCLUDING 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
 *  PURPOSE AND ANY WARRANTY AGAINST INFRINGEMENT WITH REGARD TO THE 
 *  SOFTWARE (INCLUDING ANY MODIFIED VERSIONS THEREOF) AND ANY 
 *  ACCOMPANYING WRITTEN MATERIALS.
 * 
 *  To the maximum extent permitted by applicable law, IN NO EVENT
 *  SHALL MOTOROLA BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING 
 *  WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS 
 *  INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER PECUNIARY
 *  LOSS) ARISING OF THE USE OR INABILITY TO USE THE SOFTWARE.   
 * 
 *  Motorola assumes no responsibility for the maintenance and support
 *  of this software
 ********************************************************************
 */
    .file   "reset.s"

#include "config.h"
#include "cpuio.h"
#include "warmstart.h"
#include "mcf5272.h"
    
#define STACKtop (MONSTACKSIZE-32)

    .extern   MonStack

    .global warmstart, continue
    .extern start, ledloop1, ledloop2, ledloop3

    .text

/* The tag "warmstart" is called by 'C', so put the argument
 * into reg D5.  The content of D5 must be preserved until
 * start() is called so that it can be used as the argument
 * passed into start().
 */
warmstart:
    move.l      %sp@(4),%d5

/* The tag "continue" is called by coldstart.s at the time of a
 * reset.  Refer to coldstart.s for an explanation regarding
 * why reset.s and coldstart.s are separate files.
 */
continue:
    move.w      #0x2700,%SR

    /* Invalidate and disable the cache...
	 */
    move.l  #0x01000000,%d0
    movec   %d0,%cacr

    /* Disable ACRs
	 */
    moveq.l #0,%d0
    movec   %d0,%ACR0
    movec   %d0,%ACR1

    /* Establish the base address of the internal SRAM (via RAMBAR)
	 * and set valid and C/I bits...
	 */
    move.l  #INT_SRAM_ADDRESS,%d0
    ori.l   #0x21,%d0
    movec   %d0,%RAMBAR0

    /* Initialize mcf5272 peripherals by calling
	 * mcf5272_init(IMM_ADDRESS)...
	 */
    move.l	#IMM_ADDRESS,%d0
    move.l  %d0,-(%sp)
    jsr     mcf5272_init

	/* Now that we are jumping into the monitor at C-level,
	 * the stack pointer is re-initialized to point to the
	 * top of the MonStack[] array.  Note that DRAM can only
	 * be visible once CS0 is de-activated as the global
	 * chipselect.
	 */
    move.l	#MonStack+STACKtop,%sp

	/* Transition to C...
	 */
    move.l	%d5,-(%sp)
    jsr		start

    /* Should never get here: */
badnews:
	jmp coldstart


/********************************************************************
 * Miscellaneous assembler functions:
 */
    .global putsr
    .global getsr
    .global getsp

putsr:
    move.l      %sp@(4),%d0
    move.w      %d0,%sr
    rts
    
getsr:
    move.l      #0,%d0
    move.w      %sr,%d0
    rts

getsp:
    move.l      %sp,%d0
    rts


/********************************************************************
 * mcf5272 has Icache only
 */
cpu_cache_flush:
    nop                     /* sync */
    move.l  #0x01000000,%d0 /* Invalidate the I-Cache */
    movec %d0,%cacr
    rts

/********************************************************************
 * These routines write to the special purpose registers in the ColdFire
 * core.  Since these registers are write-only in the supervisor model,
 * no corresponding read routines exist.
 */
    .global mcf5272_wr_vbr
mcf5272_wr_vbr:
    move.l  %sp@(4),%d0
    andi.l  #0xFFF00000,%d0 /* align to 1M boundary  */
    movec   %d0,%VBR
    nop
    rts

    .global mcf5272_wr_cacr
mcf5272_wr_cacr:
    move.l  %sp@(4),%d0
    movec %d0,%cacr
    nop
    rts

    .global mcf5272_wr_acr0
mcf5272_wr_acr0:
    move.l  %sp@(4),%d0
    movec %d0,%ACR0
    nop
    rts

    .global mcf5272_wr_acr1
mcf5272_wr_acr1:
    move.l  %sp@(4),%d0
    movec %d0,%ACR1
    nop
    rts

    .global mcf5272_wr_rambar
mcf5272_wr_rambar:
    move.l  %sp@(4),%d0
    movec %d0,%RAMBAR0
    nop
    rts

    .global mcf5272_wr_mbar
mcf5272_wr_mbar:
    move.l  %sp@(4),%d0
    movec %d0,%MBAR
    nop
    rts
    
    .global mcf5272_wr_sr
mcf5272_wr_sr:
    move.l  %sp@(4),%d0
    move.l  4(%sp),%d0
    move.w  %d0,%SR
    rts
