//==========================================================================
//
// csb437_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         07/31/2004
// Description:  This file contains code to intialize the CSB437 GPIO
//               section as well as set, clear and test functions
//               for manipulating the GPIO bits
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "at91rm9200.h"
#include "at91rm9200_gpio.h"
#include "csb437_gpio.h"

// Function Prototypes
void GPIO_init(void);
int GPIO_set(int);
int GPIO_clr(int);
int GPIO_tst(int);
int GPIO_in(int);
int GPIO_out(int);

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the AT91RM9200
// GPIO as used on the CSB437.  Refer to csb437_gpio.h for
// a description of the default values.  Here we just put
// them into the chip in the following order:
// 1.  Disable all GPIO pins
// 2.  Set only the pins that will be GPIO
// 3.  Set the pins that will carry alternate function A
// 4.  Set the pins that will carry alternate function B
// 5.  Disable all GPIO outputs and disable multi-drive
// 6.  Clear the GPIO output pins that need to start low
// 7.  Set the GPIO output pins that need to start high
// 8.  Disable all pullups
// 9.  Set which GPIO pins will have pullups
// 10. Enable the GPIO output only pins (I/O pins will
//     be left as inputs)

void GPIO_init()
{
	// Port A
	PIOA_REG(PIO_PDR) = 0xffffffff;							// GPIO Pin Disable
	PIOA_REG(PIO_PER) = ~(DEF_PIOA_ASR | DEF_PIOA_BSR);		// GPIO Pin Enable

	PIOA_REG(PIO_ASR) = DEF_PIOA_ASR;						// Alternate Function A
	PIOA_REG(PIO_BSR) = DEF_PIOA_BSR;						// Alternate Function B

	PIOA_REG(PIO_ODR) = 0xffffffff;							// Disable all GPIO Outputs first
	PIOA_REG(PIO_MDDR) = 0x0;								// no multidrive
	PIOA_REG(PIO_SODR) = DEF_PIOA_SODR;						// GPIO Outputs Set
	PIOA_REG(PIO_CODR) = DEF_PIOA_CODR;						// GPIO Outputs Clear

	PIOA_REG(PIO_PUDR) = 0xffffffff;						// Disable all Pullups first
	PIOA_REG(PIO_PUER) = DEF_PIOA_PUER;						// Enable Desired Pullups 

	PIOA_REG(PIO_OER) = DEF_PIOA_OER;						// GPIO Output Enable

	// Port B
	PIOB_REG(PIO_PDR) = 0xffffffff;							// GPIO Pin Disable
	PIOB_REG(PIO_PER) = ~(DEF_PIOB_ASR | DEF_PIOB_BSR);

	PIOB_REG(PIO_ASR) = DEF_PIOB_ASR;
	PIOB_REG(PIO_BSR) = DEF_PIOB_BSR;

	PIOB_REG(PIO_ODR) = 0xffffffff;
	PIOB_REG(PIO_MDDR) = 0x0;
	PIOB_REG(PIO_SODR) = DEF_PIOB_SODR;
	PIOB_REG(PIO_CODR) = DEF_PIOB_CODR;

	PIOB_REG(PIO_PUDR) = 0xffffffff;
	PIOB_REG(PIO_PUER) = DEF_PIOB_PUER;

	PIOB_REG(PIO_OER) = DEF_PIOB_OER;

	// Port C
	PIOC_REG(PIO_PDR) = 0xffffffff;							// GPIO Pin Disable
	PIOC_REG(PIO_PER) = ~(DEF_PIOC_ASR | DEF_PIOC_BSR);

	PIOC_REG(PIO_ASR) = DEF_PIOC_ASR;
	PIOC_REG(PIO_BSR) = DEF_PIOC_BSR;

	PIOC_REG(PIO_ODR) = 0xffffffff;
	PIOC_REG(PIO_MDDR) = 0x0;
	PIOC_REG(PIO_SODR) = DEF_PIOC_SODR;
	PIOC_REG(PIO_CODR) = DEF_PIOC_CODR;
	PIOC_REG(PIO_PUDR) = 0xffffffff;
	PIOC_REG(PIO_PUER) = DEF_PIOC_PUER;
	PIOC_REG(PIO_OER) = DEF_PIOC_OER;

	// Port D
	PIOD_REG(PIO_PDR) = 0xffffffff;							// GPIO Pin Disable
	PIOD_REG(PIO_PER) = ~(DEF_PIOD_ASR | DEF_PIOD_BSR);

	PIOD_REG(PIO_ASR) = DEF_PIOD_ASR;
	PIOD_REG(PIO_BSR) = DEF_PIOD_BSR;

	PIOD_REG(PIO_ODR) = 0xffffffff;
	PIOD_REG(PIO_MDDR) = 0x0;
	PIOD_REG(PIO_SODR) = DEF_PIOD_SODR;
	PIOD_REG(PIO_CODR) = DEF_PIOD_CODR;
	PIOD_REG(PIO_PUDR) = 0xffffffff;
	PIOD_REG(PIO_PUER) = DEF_PIOD_PUER;
	PIOD_REG(PIO_OER) = DEF_PIOD_OER;
}

//--------------------------------------------------------
// GPIO_set()
//
// This function sets the desired bit passed in.
// NOTE: We do not test to see if setting the bit
// would screw up any alternate functions.  Use 
// this function with caution!
//

int GPIO_set(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 128) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		PIOA_REG(PIO_SODR) = (1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		PIOB_REG(PIO_SODR) = (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		PIOC_REG(PIO_SODR) = (1 << (gpio_bit - 64));
	}
	else
	{
		// Port D
		PIOD_REG(PIO_SODR) = (1 << (gpio_bit - 96));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_clr()
//
// This function clears the desired bit passed in.
//

int GPIO_clr(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 128) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		PIOA_REG(PIO_CODR) = (1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		PIOB_REG(PIO_CODR) = (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		PIOC_REG(PIO_CODR) = (1 << (gpio_bit - 64));
	}
	else
	{
		// Port D
		PIOD_REG(PIO_CODR) = (1 << (gpio_bit - 96));
	}
	return 0;
}
//--------------------------------------------------------
// GPIO_tst()
//
// This function returns the state of desired bit passed in.
// It does not test to see if it's an input or output and thus
// can be used to verify if an output set/clr has taken place
// as well as for testing an input state.
//

int GPIO_tst(int gpio_bit)
{
	if (gpio_bit < 32)
	{
		// Port A
		if (PIOA_REG(PIO_PDSR) & (1 << (gpio_bit - 0))) return 1;
	}
	else if (gpio_bit < 64)
	{
		// Port B
		if (PIOB_REG(PIO_PDSR) & (1 << (gpio_bit - 32))) return 1;
	}
	else if (gpio_bit < 96)
	{
		// Port C
		if (PIOC_REG(PIO_PDSR) & (1 << (gpio_bit - 64))) return 1;
	}
	else
	{
		// Port D
		if (PIOD_REG(PIO_PDSR) & (1 << (gpio_bit - 96))) return 1;
	}
	return 0; // bit was not set
}

//--------------------------------------------------------
// GPIO_in()
//
// This function changes the direction of the desired bit 
// to input.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_in(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 128) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		PIOA_REG(PIO_ODR) = (1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		PIOB_REG(PIO_ODR) = (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		PIOC_REG(PIO_ODR) = (1 << (gpio_bit - 64));
	}
	else
	{
		// Port D
		PIOD_REG(PIO_ODR) = (1 << (gpio_bit - 96));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_out()
//
// This function changes the direction of the desired bit 
// to output.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_out(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 128) return -1;

	if (gpio_bit < 32)
	{
		// Port A
		PIOA_REG(PIO_OER) = (1 << (gpio_bit - 0));
	}
	else if (gpio_bit < 64)
	{
		// Port B
		PIOB_REG(PIO_OER) = (1 << (gpio_bit - 32));
	}
	else if (gpio_bit < 96)
	{
		// Port C
		PIOC_REG(PIO_OER) = (1 << (gpio_bit - 64));
	}
	else
	{
		// Port D
		PIOD_REG(PIO_OER) = (1 << (gpio_bit - 96));
	}
	return 0;
}

