/* main.c:
	General notice:
	This code is part of a boot-monitor package developed as a generic base
	platform for embedded system designs.  As such, it is likely to be
	distributed to various projects beyond the control of the original
	author.  Please notify the author of any enhancements made or bugs found
	so that all may benefit from the changes.  In addition, notification back
	to the author will allow the new user to pick up changes that may have
	been made by other users after this version of the code was distributed.

	Author:	Ed Sutter
	email:	esutter@lucent.com
	phone:	908-582-2351
*/
#include "config.h"
#include "cpu.h"
#include "cpuio.h"
#include "tfs.h"
#include "tfsprivate.h"
#include "monflags.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "warmstart.h"

ulong MonStack[0x1000];

void reinit(), init2();
extern	void SteerConsole(void);
int	StateOfMonitor, AppExitStatus, ConsoleBaudRate;

int
main(int argc,char *argv[])
{
	char ename[48];

    /* Set user level to its max, then allow monrc file to adjust it. */
    initUsrLvl(MAXUSRLEVEL);

    switch(StateOfMonitor) {
    case INITIALIZE:
    	/* Init register data structures. */
    	reginit();

		/* Second phase of hardware initialization. */
		init2();

		/* Common system initialization stuff... */
		init3();

		break;
    case MISC:
    case SSTEP:
    case BAILOUT:
    case MORESTART:
    case BREAKPOINT:
#if INCLUDE_DEBUG
		regbackup();
#endif
		break;
    case APP_EXIT:
		reinit();
    	printf("\nApplication Exit Status: %d (0x%x)\n",
			AppExitStatus,AppExitStatus);
		break;
    case EXCEPTION:
		reinit();
		printf("\nEXCEPTION: '%s'\n",ExceptionType2String(ExceptionType));
		printf("           Occurred near 0x%lx",ExceptionAddr);
		if (AddrToSym(-1,ExceptionAddr,ename,0))
			printf(" (within %s)",ename);
		printf("\n\n");
		exceptionAutoRestart(INITIALIZE);
		break;
    default:
		printf("Unexpected monitor state: 0x%x\n",StateOfMonitor);
		break;
    }	

    /* Enter the endless loop of command processing: */
    CommandLoop();
	return(0);
}

void
reinit()
{
	init1(0);		/* reinit hdware */
	init2();
#if INCLUDE_ETHERNET
	EthernetStartup(0,0);	/* reinit ethernet */
#endif
}

void
init2()
{
#if INCLUDE_FLASH
	FlashInit();		/* Init flashop data structures.  This MUST be */
						/* done prior to turning on cache!!! */
#endif

	cacheInit();		/* Initialize cache. */

#if INCLUDE_STOREMAC
	storeMac();
#endif

#if INCLUDE_ETHERNET
	enreset();	 		/* Clear the ethernet interface. */
#endif
#if INCLUDE_TFS
	tfsstartup();		/* Init tfs data structures. */
#endif
}

void
monrestart(val)
int val;
{
	warmstart(val);
}

void
intsrestore(ulong msr)
{
}

ulong
intsoff()
{
	return(0);
}

/* some stubs for now... */
void
ctxMON()
{
}

void
ctxAPP()
{
}

void
not_used()
{
	printf("WARNING: jump table entry not used\n");
}

void
appexit(int val)
{
	intsoff();
	AppExitStatus = val;
#if INCLUDE_DEBUG
	removeatpoints();
#endif
	warmstart(APP_EXIT);
}

ulong
monState()
{
	return(StateOfMonitor);
}

/* cacheInitForTarget():
 *	Called out of the generice CacheInit() function.
 *	This assigns functions to be used by TFS for I/D cache maintenance
 *	and it enables I-cache through the HID0 register (1008).
 *
 *  NOTE:
 *	The instruction cache is enabled here, so this means that we cannot
 *  jump into the 8-bit-wide bootflash without taking an exception.
 *  This means that a reburn of the monitor will cause an exception
 *	because it jumps into 8-bit space.  Not a problem, just something
 *  to be aware of.
 */
void
cacheInitForTarget()
{
}

/* extValidPassword():
	Refer to validPassword() in common/password.c for notes.
*/
int
extValidPassword(char *password, int ulvl)
{
	return(-1);
}

/* Stubs until we establish more exception handling...
 */
void
exc080(void)
{
	puts("EXCEPTION 080!\n");
	while(1);
}
void
exc100(void)
{
	puts("EXCEPTION 100!\n");
	while(1);
}
void
exc200(void)
{
	puts("EXCEPTION 200!\n");
	while(1);
}
void
exc280(void)
{
	puts("EXCEPTION 280!\n");
	while(1);
}
void
exc300(void)
{
	puts("EXCEPTION 300!\n");
	while(1);
}
void
exc380(void)
{
	puts("EXCEPTION 380!\n");
	while(1);
}
void
exc400(void)
{
	puts("EXCEPTION 400!\n");
	while(1);
}
void
exc480(void)
{
	puts("EXCEPTION 480!\n");
	while(1);
}

int
func(void)
{
	return(45);
}
