//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Luis Torrico, Cogent Computer Systems, Inc.
// Date:         05/12/2008
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to the CSB733
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "warmstart.h"
#include "mx31_uart.h"
#include "mx31.h"
#include "mx31_iomux.h"
#include "cpu_lcd.h"
#include "cpu_ipu.h"
#include "mx31_mem.h"
#include "cpu_gpio.h"
#include "fb_draw.h"
#include "umongpio.h"
#include "mx31_iomux.h"

extern ulong i2c_init(void);
extern ulong getpsr(void);
extern void putpsr(ulong);

uchar bcolor=0;		// vga black

//extern void lcd_putchar(char c);
////extern void lcd_init(void);
//extern void lcd_on(void);
//extern int lcd_color_depth;		// 4, 8 or 16
//
//// globals to hold background and foreground colors
//uchar bcolor=0;		// vga black
//uchar color=15;		// vga white

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  volatile int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}


/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
    if (baud == 0)
        baud = ConsoleBaudRate;

    /* Keep track of the last baud rate, so that it can be used if the */
    /* incoming baudrate is NULL. */
    ConsoleBaudRate = baud;

	// reset tx and rx
	UART4_REG(UART_CR2) = 0;

	// clear any pending interrupts and mask all further interrupts
	UART4_REG(UART_SR1) = 0xffffffff;
	UART4_REG(UART_SR2) = 0xffffffff;
	UART4_REG(UART_CR1) = 0x00000000;
	UART4_REG(UART_CR4) = 0x00000000;

	// set the RX mux to Uart, not IrDA - this is different than i.MXL
	UART4_REG(UART_CR3) = UART_CR3_RXDMUXSEL;

	// enable the uart and it's clocks - no auto baud, no interrupts
	UART4_REG(UART_CR1) |= UART_CR1_UARTEN;

	//enable UART1 and UART2 clocks
	CCM_REG(CCM_CGR0) |= 0x00F00000;
	//enable UART3, UART4 and UART5 clocks
	CCM_REG(CCM_CGR1) |= 0x3F000000;

	// set the word parameters
	UART4_REG(UART_CR2) = UART_CR2_WS 	// 8-bit, 1 stop, no parity
					| UART_CR2_IRTS		// ignore RTS 
					| UART_CR2_SRST;	// no reset

	// set the fifo's to legal interrupt threshold values - unused by us
	// set input clock divisor to 2
	UART4_REG(UART_FCR) = UART_FCR_TXTL(2)
					| UART_FCR_RFDIV2
					| UART_FCR_RXTL(1);

	// set the baud rate denominator to 10,500 (this value used
	// to be 10,000 for MXL and MX21 boards. For MX31 boards this
	// needs to change to 10,000 b/c Pre-multiplier has gone from
	// 512 to 1024.  Then divide the desired baud by 100 and put
	// that into the numerator.  This works when using a reference
	// clock of 33.34Mhz with RFDIV set to 2.  It also only works
	// down to 1200 baud, but who cares!
	UART4_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART4_REG(UART_BMR) = 10500 - 1;		// denominator

	// Enable TX and RX, release SRST
	UART4_REG(UART_CR2) |= UART_CR2_TXEN
					| UART_CR2_RXEN
					| UART_CR2_SRST;

    InitRemoteIO();
}

int
ConsoleBaudSet(int baud)
{
	// set the baud rate denominator to 10,500 (this value used
	// to be 10,000 for MXL and MX21 boards. For MX31 boards this
	// needs to change to 10,000 b/c Pre-multiplier has gone from
	// 512 to 1024.  Then divide the desired baud by 100 and put
	// that into the numerator.  This works when using a reference
	// clock of 33.34Mhz with RFDIV set to 2.  It also only works
	// down to 1200 baud, but who cares!
	UART4_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART4_REG(UART_BMR) = 10500 - 1;		// denominator
	return(0);
}



int
devInit(int baud)
{
	// Set CLK_OUT to ipg_clk_ipu_hsp for debugging
//	CCM_REG(CCM_COSR) = 0x0000024a;
	
	// enable all perpihperal clocks except EMI.  They have to enable themselves
	CCM_REG(CCM_CGR0) = 0x00f0c000;
	CCM_REG(CCM_CGR1) = 0x3fc00000;
	CCM_REG(CCM_CGR2) = 0xffffffff;

	// Initialize IO mux
	iomux_init();

	// Initialize GPIO
    GPIO_init();	

    // Initialize the UART
    InitUART(baud);

	// Setup CS0 for 110ns Spansion Flash
    WEIM_REG(WEIM_CS0U) = WEIM_CSU_BCD(0)		// Burst Clock Divisor
						| WEIM_CSU_BCS(0)		// Burst Clock Start
						| WEIM_CSU_PSZ_8
						| WEIM_CSU_DOL(0)		// Data Output Length
						| WEIM_CSU_CNC_0		// Hold CS negated after end of cycle for 0 clocks
						| WEIM_CSU_WSC(16)		// wait = 200ns (16+1 x 12ns)
						| WEIM_CSU_WWS(0)		// additional wait states for writes
						| WEIM_CSU_EDC(0);		// 1 clock bus turn around time

	WEIM_REG(WEIM_CS0L) = WEIM_CSL_OEA(0)		// CS asserted # 1/2 clocks before OE
						| WEIM_CSL_OEN(0)		// CS held # 1/2 clocks after OE
						| WEIM_CSL_WEA(0)		// CS asserted # 1/2 clocks before WE
						| WEIM_CSL_WEN(0)		// CS held # 1/2 clocks after WE
						| WEIM_CSL_CSA(0)		// setup and hold address 2 clocks around CS
						| WEIM_CSL_DSZ_HALF_0	// 16-bits on D0-15
						| WEIM_CSL_CSEN;		// enable it

	// Setup CS1 for LAN9211
    WEIM_REG(WEIM_CS1U) = WEIM_CSU_CNC_3		//
						| WEIM_CSU_WSC(24)		//
						| WEIM_CSU_WWS(4)		// 
						| WEIM_CSU_EDC(3);		// 

	WEIM_REG(WEIM_CS1L) = WEIM_CSL_OEA(1)		// CS asserted # 1/2 clocks before OE
			  			| WEIM_CSL_OEN(0)		// CS held # 1/2 clocks after OE
			  			| WEIM_CSL_WEA(0)		// CS asserted # 1/2 clocks before WE
			  			| WEIM_CSL_WEN(0)		// CS held # 1/2 clocks after WE
			  			| WEIM_CSL_CSA(0)		// setup and hold address 2 clocks around CS
			  			| WEIM_CSL_DSZ_HALF_0	// 16-bits on D0-15
						| WEIM_CSL_CSN(0)
						| WEIM_CSL_CSEN;		// enable it

	WEIM_REG(WEIM_CS1A) = WEIM_CSA_RWA(7)		// 
						| WEIM_CSA_RWN(2)
						| WEIM_CSA_LBN(2)
						| WEIM_CSA_LBA_1;

	// Initialize the I2C Port
	//i2c_init();

	// Turn on USR_LED (GPIO 0)
	GPIO_clr(GPIO0);
	GPIO_set(GPIO0);
	GPIO_clr(GPIO0);

	return 0;
}

int
target_console_empty(void)
{
	if (UART4_REG(UART_SR2) & UART_SR2_TXFE)
		return(1);
	return(0);
}

int
target_putchar(char c)
{
    /* Wait for transmit ready bit, then send character.
	 */
    while(!(UART4_REG(UART_SR2) & UART_SR2_TXFE)){}
	UART4_REG(UART_TXD) = (ulong)c;
    return((int)c);
}

int
target_gotachar(void)
{
	if (UART4_REG(UART_SR2) & UART_SR2_RDR)
        return(1);
    return(0);
}

int 
target_getchar(void)
{
    char    c;

    c = (uchar)UART4_REG(UART_RXD) & 0xff;
    return((int)c);
}

void
initCPUio()
{
}

void 
target_reset(void)
{
	warmstart(INITIALIZE);
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff(void)
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* target_timer():
 * Used in conjunction with INCLUDE_HWTMR and TIMER_TICKS_PER_MSEC
 * to set up a hardware based time base.
 */
unsigned long
target_timer(void)
{
	volatile unsigned register ccr;

   	asm("   MRC p15, 0, %0, c9, c13, 0" : "=r" (ccr));

	return(ccr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}
