#include "config.h"
#include "stddefs.h"
#include "cpuio.h"
#include "genlib.h"
#include "cache.h"
#include "warmstart.h"
#include "timer.h"
#include "mcf532x.h"

extern void lcd_putchar(char c);
extern void lcd_init(void);
extern void lcd_on(void);
extern int lcd_color_depth;		// 4, 8 or 16

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  volatile int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

/* devInit():
 * As a bare minimum, initialize the console UART here using the
 * incoming 'baud' value as the baud rate.
 */
void
devInit(int baud)
{
	int divisor;
	uint32_t clock_speed;

	/* Pin assignments for port TIMER.  We set this port up because CSB760 uses UART2 as console port
			Pin TIMER3 : UART 2 receive data, U2RXD 
			Pin TIMER2 : UART 2 transmit data, U2TXD 
			Pin TIMER1 : GPIO3 
			Pin TIMER0 : GPIO2
	*/
	
	/* Pin assignments for port UART 
			Pin UART7 : UART 1 clear-to-send, /U1CTS 
			Pin UART6 : UART 1 request-to-send, /U1RTS 
			Pin UART5 : UART 1 transmit data, U1TXD 
			Pin UART4 : UART 1 receive data, U1RXD 
			Pin UART3 : UART 0 clear-to-send, /U0CTS 
			Pin UART2 : UART 0 request-to-send, /U0RTS 
			Pin UART1 : UART 0 transmit data, U0TXD 
			Pin UART0 : UART 0 receive data, U0RXD 
	*/
	MCF_GPIO_PDDR_TIMER = 0;
	MCF_GPIO_PAR_TIMER = MCF_GPIO_PAR_TIMER_PAR_TIN3_URXD2 |
						MCF_GPIO_PAR_TIMER_PAR_TIN2_UTXD2 |
						MCF_GPIO_PAR_TIMER_PAR_TIN1_GPIO |
						MCF_GPIO_PAR_TIMER_PAR_TIN0_GPIO;

	MCF_GPIO_PDDR_UART = 0;
	MCF_GPIO_PAR_UART = MCF_GPIO_PAR_UART_PAR_UCTS0 |
						MCF_GPIO_PAR_UART_PAR_URTS0 |
						MCF_GPIO_PAR_UART_PAR_URXD0 |
						MCF_GPIO_PAR_UART_PAR_UTXD0;
	
	/* Configure UART 2 (third UART channel) 
		Baud Rate = 38400 
		8 data bits, 1 stop bit, No parity 
		/CTS input has no effect on the transmitter 
		/RTS output is not affected by receiver or transmitter 
	*/
	
	/* UART 2: Reset transmitter, receiver and mode register pointer */
	MCF_UART2_UCR = MCF_UART_UCR_MISC(0x3);
	MCF_UART2_UCR = MCF_UART_UCR_MISC(0x2);
	MCF_UART2_UCR = MCF_UART_UCR_MISC(0x1);
	
	/* UART 2 Mode Registers */
	MCF_UART2_UMR = MCF_UART_UMR_PM(0x2) |
					MCF_UART_UMR_BC(0x3);
	MCF_UART2_UMR = MCF_UART_UMR_SB(0x7);
	
	/* UART 2 Clocking */
	clock_speed = BUS_FREQUENCY;
	/* determine the baud divisor value */
	divisor = ((clock_speed) / (32 * baud));
	if (divisor < 2)
		divisor = 2;
	
	MCF_UART2_UCSR = MCF_UART_UCSR_RCS(0xd) |
						MCF_UART_UCSR_TCS(0xd);
	MCF_UART2_UBG1 = (divisor & 0xff00) >> 8;
	MCF_UART2_UBG2 = (divisor & 0x00ff);
	
	/* UART 2 Interrupt mask */
	MCF_UART2_UIMR = 0;
	
	/* UART 2 Miscellaneous */
	MCF_UART2_UACR = 0;
	
	/* UART 2: Enable receiver and transmitter */
	MCF_UART2_UCR = MCF_UART_UCR_TXC(0x1);
	MCF_UART2_UCR = MCF_UART_UCR_RXC(0x1);

	// Initialize the LCD controller
 	lcd_color_depth = 16;		// 4, 8 or 16
	//lcd_init();
	//lcd_on();
}

/* ConsoleBaudSet():
 * Provide a means to change the baud rate of the running
 * console interface.  If the incoming value is not a valid
 * baud rate, then default to 9600.
 * In the early stages of a new port this can be left empty.
 * Return 0 if successful; else -1.
 */
int
ConsoleBaudSet(int baud)
{
	int divisor;
	uint32_t clock_speed;

	/* disable interrupts, clear RTS line, and disable the UARTS */
	MCF_UART2_UIMR = 0;
	MCF_UART2_UOP0 = 1;
	MCF_UART2_UCR = (MCF_UART_UCR_TX_DISABLED | MCF_UART_UCR_RX_DISABLED);

	/* UART 2 Clocking */
	clock_speed = BUS_FREQUENCY;
	/* determine the baud divisor value */
	divisor = ((clock_speed) / (32 * baud));
	if (divisor < 2)
		divisor = 2;
	
	/* reset the uart */
	MCF_UART2_UCR = MCF_UART_UCR_RESET_ERROR;
	MCF_UART2_UCR = MCF_UART_UCR_RESET_RX;
	MCF_UART2_UCR = MCF_UART_UCR_RESET_TX;

	/* reset the uart mode register and update values */
	MCF_UART2_UCR = MCF_UART_UCR_RESET_MR;
	MCF_UART2_UMR = MCF_UART_UMR_PM(0x2) |
					MCF_UART_UMR_BC(0x3);
	MCF_UART2_UMR = MCF_UART_UMR_SB(0x7);

	MCF_UART2_UCSR = MCF_UART_UCSR_RCS(0xd) |
						MCF_UART_UCSR_TCS(0xd);
	MCF_UART2_UBG1 = (divisor & 0xff00) >> 8;
	MCF_UART2_UBG2 = (divisor & 0x00ff);
	
	/* UART 2 Interrupt mask */
	MCF_UART2_UIMR = 0;
	
	/* UART 2 Miscellaneous */
	MCF_UART2_UACR = 0;
	
	/* UART 2: Enable receiver and transmitter */
	MCF_UART2_UCR = MCF_UART_UCR_TXC(0x1);
	MCF_UART2_UCR = MCF_UART_UCR_RXC(0x1);

	return(0);
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 * In the early stages of a new port this can simply return 1.
 */
int
target_console_empty(void)
{
	return (MCF_UART2_USR & MCF_UART_USR_TXRDY);
}

/* target_putchar():
 * When buffer has space available, load the incoming character
 * into the UART.
 */
int
target_putchar(char c)
{
	/* Wait until space is available in the FIFO */
	while (!(MCF_UART2_USR & MCF_UART_USR_TXRDY)) ;
	
	/* Send the character */
	MCF_UART2_UTB = (unsigned char) c;
	//lcd_putchar(c);
	return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 *
 * Define INCLUDE_BLINKLED in config.h and add STATLED_ON()
 * and STATLED_OFF() macros (or functions)  so that uMon will
 * run and blink a target LED at a configured (or default 500msec)
 * interval...
 */
int
target_gotachar(void)
{
#if INCLUDE_BLINKLED
	static uint8_t ledstate;
	static struct elapsed_tmr tmr;

#ifndef BLINKON_MSEC
#define BLINKON_MSEC 500
#define BLINKOFF_MSEC 500
#endif

	switch(ledstate) {
		case 0:
			startElapsedTimer(&tmr,BLINKON_MSEC);
			ledstate = 1;
			STATLED_ON();
			break;
		case 1:
			if(msecElapsed(&tmr)) {
				STATLED_OFF();
				ledstate = 2;
				startElapsedTimer(&tmr,BLINKOFF_MSEC);
			}
			break;
		case 2:
			if(msecElapsed(&tmr)) {
				STATLED_ON();
				ledstate = 1;
				startElapsedTimer(&tmr,BLINKON_MSEC);
			}
			break;
	}
#endif

	return (MCF_UART2_USR & MCF_UART_USR_RXRDY);
}

/* target_getchar():
 * Assume there is a character in the UART's input buffer and just
 * pull it out and return it.
 */
int 
target_getchar(void)
{
	/* Wait until character has been received */
	return MCF_UART2_URB;
}

/* intsoff():
 * Disable all system interrupts here and return a value that can
 * be used by intsrestore() (later) to restore the interrupt state.
 */
ulong
intsoff(void)
{
	extern void putsr(ulong);
	extern ulong getsr(void);
	ushort  sr;

	sr = getsr();
	putsr(sr | 0x0700);
	return((int) sr);
}

/* intsrestore():
 * Re-establish system interrupts here by using the status value
 * that was returned by an earlier call to intsoff().
 */
void
intsrestore(ulong status)
{
	extern void putsr(ulong);
	putsr(status);
}

/* The 5329 has a unified data and instruction cache 
 */

/* uninitialized shadow copy of CACR */
unsigned long cacrShadow;

static inline void
targetCacheFlush1Line(const void *adr)
{
	register unsigned long set = (((unsigned long) adr >> 4) & 0xff) << 4;

	asm volatile ("cpushl %%bc,(%0)" :: "a" (set));
	set += 1;
	asm volatile ("cpushl %%bc,(%0)" :: "a" (set));
	set += 1;
	asm volatile ("cpushl %%bc,(%0)" :: "a" (set));
	set += 1;
	asm volatile ("cpushl %%bc,(%0)" :: "a" (set));
}


int
targetDcacheFlush(char * adr, int n)
{
	if(adr == 0 && n == 0)
	{
		int set;
		for(set = 0; set < 256; ++set)
		{
			targetCacheFlush1Line((const void *) (set << 4));
		}
	}
	else if(n)
	{
		char* final_address = (void *) ((unsigned long) adr + n - 1);
		adr = (char *) (((unsigned long) adr) & ~15);
		while(adr <= final_address)
		{
			targetCacheFlush1Line(adr);
			adr += 16;
		}
	}
	return 0;
}

int 
targetIcacheInvalidate(char * adr, int n)
{
	if(adr == 0 && n == 0)
	{
		asm volatile ("nop");
		asm volatile ("movec %0,%%cacr" :: "d" (cacrShadow | MCF_CACR_CINVA));
	}
	else if(n)
	{
		char* final_address = (void *) ((unsigned long) adr + n - 1);
		adr = (char *) (((unsigned long) adr) & ~15);
		while(adr <= final_address)
		{
			targetCacheFlush1Line(adr);
			adr += 16;
		}
	}
	return 0;
}

int
targetDcacheDisable(void)
{
	unsigned long sr = intsoff();
	cacrShadow &= ~MCF_CACR_CENB;
	asm volatile ("movec %0,%%cacr" :: "d" (cacrShadow | MCF_CACR_CINVA));
	intsrestore(sr);
	return 0;
}

int
targetIcacheDisable(void)
{
	unsigned long sr = intsoff();
	cacrShadow &= ~MCF_CACR_CENB;
	asm volatile ("movec %0,%%cacr" :: "d" (cacrShadow | MCF_CACR_CINVA));
	intsrestore(sr);
	return 0;
}


/* cacheInitForTarget():
 * Establish target specific function pointers and
 * enable i-cache...
 * Refer to $core/cache.c for a description of the function pointers.
 * NOTE:
 * If cache (either I or D or both) is enabled, then it is important
 * that the appropriate cacheflush/invalidate function be established.
 * This is very important because programs (i.e. cpu instructions) are
 * transferred to memory using data memory accesses and could
 * potentially result in cache coherency problems.
 */
void
cacheInitForTarget(void)
{
#if defined(DONT_USE_SRAM) || defined(DONT_USE_CACHE)
	dcacheFlush = 0;
	icacheInvalidate = 0;
	dcacheDisable = 0;
	icacheDisable = 0;

	/* Disable the cache */
    asm("move.l   #0x00000000,%d0");
    asm("movec    %d0,%CACR");
#else
	dcacheFlush = targetDcacheFlush;
	icacheInvalidate = targetIcacheInvalidate;
	dcacheDisable = targetDcacheDisable;
	icacheDisable = targetIcacheDisable;

	/* Enable the cache.  Note that the cache has already been 
	 * invalidated in the warmstartup */
	cacrShadow = MCF_CACR_ESB | MCF_CACR_DCM(3) | MCF_CACR_CENB;
	asm volatile ("movec %0,%%cacr" :: "d" (cacrShadow));

	/* ACR0: Cache accesses to 32 MB memory region at address $40000000
	 * Copyback
	 */
	asm("move.l   #0x4001c020,%d0");
	asm("movec    %d0,%ACR0");

	/* ACR0: Cache accesses to 16 MB memory region at address $00000000 
	 * Writethrough
	 */
    //asm("move.l   #0x0000c000,%d0");
    //asm("movec    %d0,%ACR1");
#endif
}

/* target_reset():
 * The default (absolute minimum) action to be taken by this function
 * is to call monrestart(INITIALIZE).  It would be better if there was
 * some target-specific function that would really cause the target
 * to reset...
 */
void
target_reset(void)
{
	flushDcache(0,0);
	disableDcache();
	invalidateIcache(0,0);
	disableIcache();

	/*
	 * Cause a soft reset, we would otherwise call
	 * monrestart(INITIALIZE);
	 */
	MCF_RCM_RCR = MCF_RCM_RCR_SOFTRST;
}

/* If any CPU IO wasn't initialized in reset.S, do it here...
 * This just provides a "C-level" IO init opportunity. 
 */
void
initCPUio(void)
{
	/* The function init_main() has already been called by warmstart
	 * so there's nothing to do here. 
	 */
}
